#!/usr/bin/python -OO
# Copyright 2005 Gregor Kaufmann <tdn@gmx.at>
# Licensed under the Academic Free License 2.1

__version__ = "0.1.8.1"
__configversion__ = 11
__queueversion__ = 1
__NAME__ = "sabnzbd"

import os
import logging
import datetime

from threading import Lock, Condition, currentThread

from sabnzbd.assembler import Assembler, PostProcessor
from sabnzbd.cron import CronThread
from sabnzbd.downloader import Downloader, BPSMeter
from sabnzbd.nzbqueue import NzbQueue
from sabnzbd.misc import MSGIDGrabber, URLGrabber, DirScanner
from sabnzbd.nzbstuff import NzbObject

LOCK = Lock()

CV = Condition()

START = datetime.datetime.now()

CFG = None

RAM_CACHE = False
FAIL_ON_CRC = False
CREATE_GROUP_FOLDERS = False
CONSERVE_MEMORY = False
DO_FILE_JOIN = False
DO_UNZIP = False
DO_UNRAR = False
DO_SAVE = False
PAR_CLEANUP = False
CLEANUP_LIST = None

USERNAME_NEWZBIN = None
PASSWORD_NEWZBIN = None

NZB_BACKUP_DIR = None
DOWNLOAD_DIR = None
COMPLETE_DIR = None

POSTPROCESSOR = None
ASSEMBLER = None
DIRSCANNER = None
CT = None
NZBQ = None
DOWNLOADER = None

URLGRABBERS = []
MSGIDGRABBERS = []

__INITIALIZED__ = False

################################################################################
# Decorators                                                                   #
################################################################################
def synchronized(func):
    def call_func(*params, **kparams):
        LOCK.acquire()
        try:
            return func(*params, **kparams)
        finally:
            LOCK.release()
    return call_func
    
def synchronized_CV(func):
    def call_func(*params, **kparams):
        CV.acquire()
        try:
            return func(*params, **kparams)
        finally:
            CV.notifyAll()
            CV.release()
    return call_func
    
################################################################################
# Initializing                                                                 #
################################################################################
def initialize():
    global __INITIALIZED__, RAM_CACHE, FAIL_ON_CRC, CREATE_GROUP_FOLDERS, \
           CONSERVE_MEMORY, DO_FILE_JOIN, DO_UNZIP, DO_UNRAR, DO_SAVE, \
           PAR_CLEANUP, CLEANUP_LIST, USERNAME_NEWZBIN, PASSWORD_NEWZBIN, \
           POSTPROCESSOR, ASSEMBLER, DIRSCANNER, CT, NZBQ, DOWNLOADER, \
           NZB_BACKUP_DIR, DOWNLOAD_DIR, COMPLETE_DIR, BYTES_FILE, TOTAL_BYTES
           
    if __INITIALIZED__:
        return False
           
    logging.info("Initializing SABnzbd v%s", __version__)
    
    ###########################
    ## CONFIG Initialization ##
    ###########################
    
    USERNAME_NEWZBIN = CFG['newzbin']['username']
    PASSWORD_NEWZBIN = CFG['newzbin']['password'] 
    
    RAM_CACHE = bool(int(CFG['misc']['ram_cache']))
    logging.debug("RAM_CACHE -> %s", RAM_CACHE)
    
    FAIL_ON_CRC = bool(int(CFG['misc']['fail_on_crc']))
    logging.debug("FAIL_ON_CRC -> %s", FAIL_ON_CRC)
    
    CREATE_GROUP_FOLDERS = bool(int(CFG['misc']['create_group_folders']))
    logging.debug("CREATE_GROUP_FOLDERS -> %s", CREATE_GROUP_FOLDERS)
    
    DO_FILE_JOIN = bool(int(CFG['misc']['enable_filejoin']))
    logging.debug("DO_FILE_JOIN -> %s", DO_FILE_JOIN)
    
    DO_UNZIP = bool(int(CFG['misc']['enable_unzip']))
    logging.debug("DO_UNZIP -> %s", DO_UNZIP)
    
    DO_UNRAR = bool(int(CFG['misc']['enable_unrar']))
    logging.debug("DO_UNRAR -> %s", DO_UNRAR)
    
    DO_SAVE = bool(int(CFG['misc']['enable_save']))
    logging.debug("DO_SAVE -> %s", DO_SAVE)
    
    CONSERVE_MEMORY = bool(int(CFG['misc']['conserve_memory']))
    logging.debug("CONSERVE_MEMORY -> %s", CONSERVE_MEMORY)
    
    PAR_CLEANUP = bool(int(CFG['misc']['enable_par_cleanup']))
    logging.debug("PAR_CLEANUP -> %s", PAR_CLEANUP)
    
    CLEANUP_LIST = CFG['misc']['cleanup_list']
    logging.debug("CLEANUP_LIST -> %s", CLEANUP_LIST)
    
    if not CFG['misc']['download_dir']:
        logging.error('No DOWNLOAD_DIR defined!')
        return False
    
    DOWNLOAD_DIR = os.path.abspath(CFG['misc']['download_dir'])
    if not os.path.exists(DOWNLOAD_DIR):
        logging.error('Download directory: %s does not exist', DOWNLOAD_DIR)
        return False
    if not os.access(DOWNLOAD_DIR, os.R_OK + os.W_OK):
        logging.error('Download directory: %s error accessing',
                      DOWNLOAD_DIR)
        return False
    logging.info("DOWNLOAD_DIR: %s", DOWNLOAD_DIR)
    
    COMPLETE_DIR = CFG['misc']['complete_dir']
    if COMPLETE_DIR:
        COMPLETE_DIR = os.path.abspath(COMPLETE_DIR)
        if not os.path.exists(COMPLETE_DIR):
            logging.error('Directory: %s does not exist', COMPLETE_DIR)
            return False
        if not os.access(COMPLETE_DIR, os.R_OK + os.W_OK):
            logging.error('Directory: %s error accessing', COMPLETE_DIR)
            return False
    logging.info("COMPLETE_DIR: %s", COMPLETE_DIR)
    
    if "samefile" in os.path.__dict__:
        if os.path.samefile(DOWNLOAD_DIR, COMPLETE_DIR):
            logging.error('DOWNLOAD_DIR and COMPLETE_DIR cannot be the same!')
            return True
        
    if not CFG['misc']['cache_dir']:
        logging.error('No cache_dir defined!')
        return False
    
    cache_dir_root = os.path.abspath(CFG['misc']['cache_dir'])
    if not os.path.exists(cache_dir_root):
        logging.error('Root Cache directory directory: %s does not exist', 
                      cache_dir_root)
        return False
    if not os.access(cache_dir_root, os.R_OK + os.W_OK):
        logging.error('Root Cache directory directory: %s error accessing', 
                      cache_dir_root)
        return False
    logging.info("cache_dir_root: %s", cache_dir_root)
    
    
    dirscan_dir = CFG['misc']['dirscan_dir']
    if dirscan_dir:
        dirscan_dir = os.path.abspath(dirscan_dir)
        if not os.path.exists(dirscan_dir):
            logging.error('Directory: %s does not exist', dirscan_dir)
            return False
        if not os.access(dirscan_dir, os.R_OK + os.W_OK):
            logging.error('Directory: %s error accessing', dirscan_dir)
            return False
    logging.info("dirscan_dir: %s", dirscan_dir)
    
    NZB_BACKUP_DIR = CFG['misc']['nzb_backup_dir']
    if NZB_BACKUP_DIR:
        NZB_BACKUP_DIR = os.path.abspath(NZB_BACKUP_DIR)
        if not os.path.exists(NZB_BACKUP_DIR):
            logging.error('Directory: %s does not exist', NZB_BACKUP_DIR)
            return False
        if not os.access(NZB_BACKUP_DIR, os.R_OK + os.W_OK):
            logging.error('Directory: %s error accessing', NZB_BACKUP_DIR)
            return False
    logging.info("NZB_BACKUP_DIR: %s", NZB_BACKUP_DIR)
    
    cache_dir_articles = os.path.abspath(os.path.join(cache_dir_root, 'cache_articles'))
    cache_dir_items = os.path.abspath(os.path.join(cache_dir_root, 'cache_items'))
    cache_file = os.path.abspath(os.path.join(cache_dir_root, 'SABnzbd.qcache'))
    
    logging.info("cache_dir_articles: %s", cache_dir_articles)
    logging.info("cache_dir_items: %s", cache_dir_items)
    logging.info("cache_file: %s", cache_file)
    
    servers = CFG['servers']
    
    bw_limit = float(CFG['misc']['bandwith_limit'])
    logging.info("bw_limit: %s", bw_limit)
    
    cronlines = CFG['misc']['cronlines']
    logging.info("cronlines: %s", cronlines)
    
    dirscan_opts = int(CFG['misc']['dirscan_opts'])
    dirscan_repair, dirscan_unpack, dirscan_delete = pp_to_opts(dirscan_opts)
    logging.info("dirscan_opts: %s", dirscan_opts)
    
    f_mode = bool(int(CFG['misc']['f_mode']))
    logging.info("f_mode: %s", f_mode)
    
    auto_sort = bool(int(CFG['misc']['auto_sort']))
    logging.info("auto_sort: %s", auto_sort)
    
    ##############################
    ## Directory initialization ##
    ##############################
    if not os.path.exists(cache_dir_articles):
        try:
            os.mkdir(cache_dir_articles, 0744)
        except OSError:
            logging.exception("Error creating %s", cache_dir_articles)
            return False
    
    if not os.path.exists(cache_dir_items):
        try:
            os.mkdir(cache_dir_items, 0744)
        except OSError:
            logging.exception("Error creating %s", cache_dir_items)
            return False
    
    ############################
    ## Object initializiation ##
    ############################
    if NZBQ:
        NZBQ.__init__(cache_file, cache_dir_items, cache_dir_articles, f_mode,
                      auto_sort)
    else:
        NZBQ = NzbQueue(cache_file, cache_dir_items, cache_dir_articles, f_mode,
                        auto_sort)
        
    if POSTPROCESSOR:
        POSTPROCESSOR.__init__(DOWNLOAD_DIR, COMPLETE_DIR, POSTPROCESSOR.queue)
    else:
        POSTPROCESSOR = PostProcessor(DOWNLOAD_DIR, COMPLETE_DIR)
         
    if ASSEMBLER:
        ASSEMBLER.__init__(DOWNLOAD_DIR, cache_dir_articles, ASSEMBLER.queue)
    else:
        ASSEMBLER = Assembler(DOWNLOAD_DIR, cache_dir_articles)
        
    if DOWNLOADER:
        DOWNLOADER.__init__(servers, bw_limit, DOWNLOADER.paused)
    else:
        DOWNLOADER = Downloader(servers, bw_limit)
        
    if cronlines:
        CT = CronThread(cronlines)
        
    if dirscan_dir:
        DIRSCANNER = DirScanner(dirscan_dir, dirscan_repair, dirscan_unpack, 
                                dirscan_delete)
                                
    __INITIALIZED__ = True
    return True
    
def start():
    if __INITIALIZED__:
        logging.debug('[%s] Starting downloader', __NAME__)
        POSTPROCESSOR.start()
            
        logging.debug('[%s] Starting assembler', __NAME__)
        ASSEMBLER.start()
            
        logging.debug('[%s] Starting downloader', __NAME__)
        DOWNLOADER.start()
            
        if CT:
            logging.debug('[%s] Starting ct', __NAME__)
            CT.start()
            
        if DIRSCANNER:
            logging.debug('[%s] Starting dirscanner', __NAME__)
            DIRSCANNER.start()
            
def halt():
    global __INITIALIZED__, CT, DIRSCANNER
    
    if __INITIALIZED__:
        logging.info('SABnzbd shutting down...')
        
        ## Stop Optional Objects ##
        for grabber in URLGRABBERS:
            logging.debug('Stopping grabber {%s}', grabber)
            try:
                grabber.join()
            except:
                pass
                
        for grabber in MSGIDGRABBERS:
            logging.debug('Stopping grabber {%s}', grabber)
            try:
                grabber.join()
            except:
                pass
                
        if DIRSCANNER:
            logging.debug('Stopping dirscanner')
            DIRSCANNER.stop()
            try:
                DIRSCANNER.join()
            except:
                pass
            DIRSCANNER = None
                
        if CT:
            logging.debug('Stopping ct')
            CT.stop()
            try:
                CT.join()
            except:
                pass
            CT = None
        
        ## Stop Required Objects ##
        logging.debug('Stopping downloader')
        CV.acquire()
        try:
            DOWNLOADER.stop()
        finally:
            CV.notifyAll()
            CV.release()
        try:
            DOWNLOADER.join()
        except:
            pass
            
        logging.debug('Stopping assembler')
        ASSEMBLER.stop()
        try:
            ASSEMBLER.join()
        except:
            pass
            
        logging.debug('Stopping postprocessor')
        POSTPROCESSOR.stop()
        try:
            POSTPROCESSOR.join()
        except:
            pass
        
        ## Save State ##
        NZBQ.save()
            
        __INITIALIZED__ = False
        
################################################################################
## Synchronized Methods                                                       ##
################################################################################
@synchronized
def get_article(host, fillserver):
    try:
        return NZBQ.get_article(host, fillserver)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
            
@synchronized
def has_articles():
    try:
        return not NZBQ.is_empty()
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
        
@synchronized
def has_articles_for(host):
    try:
        return not (host in NZBQ.try_list)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
            
@synchronized
def register_article(article, lines = None):
    try:
        return NZBQ.register_article(article, lines)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
            
@synchronized
def remove_from_try_list(host):
    try:
        NZBQ.remove_from_try_list(host)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
            
@synchronized
def remove_nzo(nzo_id, add_to_history = True):
    try:
        NZBQ.remove(nzo_id, add_to_history)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
            
@synchronized
def cleanup_nzo(nzo):
    try:
        NZBQ.cleanup_nzo(nzo)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
            
@synchronized
def remove_nzf(nzo_id, nzf_id):
    try:
        NZBQ.remove_nzf(nzo_id, nzf_id)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
        
@synchronized
def sort_by_avg_age():
    try:
        NZBQ.sort_by_avg_age()
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
        
@synchronized
def add_msgid(msgid, pp):
    logging.info('[%s] Fetching id %s from www.newzbin.com',
                 __NAME__, id)
    msg = "fetching msgid %s from www.newzbin.com" % msgid
    
    repair, unpack, delete = pp_to_opts(pp)
    
    future_nzo = NZBQ.generate_future(msgid, msg, repair, unpack, delete)
    
    # Look for a grabber and reinitialize it
    for grabber in MSGIDGRABBERS:
        if not grabber.isAlive():
            grabber.__init__(USERNAME_NEWZBIN, PASSWORD_NEWZBIN, future_nzo)
            grabber.start()
            return
            
    grabber = MSGIDGrabber(USERNAME_NEWZBIN, PASSWORD_NEWZBIN, future_nzo)
    grabber.start()
    
    MSGIDGRABBERS.append(grabber)
    
@synchronized
def add_url(url, pp):
    logging.info('[%s] Fetching %s', __NAME__, url)
    
    msg = "Trying to fetch .nzb from %s" % url
    
    repair, unpack, delete = pp_to_opts(pp)
    
    future_nzo = NZBQ.generate_future(None, msg, repair, unpack, delete)
    
    # Look for a grabber and reinitialize it
    for urlgrabber in URLGRABBERS:
        if not urlgrabber.isAlive():
            urlgrabber.__init__(url, future_nzo)
            urlgrabber.start()
            return
            
    urlgrabber = URLGrabber(url, future_nzo)
    urlgrabber.start()
    
    URLGRABBERS.append(urlgrabber)
        
@synchronized
def switch(nzo_id1, nzo_id2):
    try:
        NZBQ.switch(nzo_id1, nzo_id2)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
            
@synchronized
def history_info():
    try:
        return NZBQ.history_info()
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
            
@synchronized
def queue_info():
    try:
        return NZBQ.queue_info()
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
            
@synchronized
def purge_history():
    try:
        NZBQ.purge()
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)

# Makes sure that only one thread writes to a filename
@synchronized
def backup_nzb(filename, data):
    if NZB_BACKUP_DIR:
        try:
            path = os.path.join(NZB_BACKUP_DIR, filename)
            logging.info("[%s] Saving %s", __NAME__, path)
            _f = open(path, 'w')
            _f.write(data)
            _f.flush()
            _f.close()
        except:
            logging.exception("[%s] Saving %s failed", __NAME__, path)
            
################################################################################
## CV synchronized (notifys downloader)                                       ##
################################################################################
@synchronized_CV
@synchronized
def add_nzbfile(nzbfile, pp):
    
    repair, unpack, delete = pp_to_opts(pp)
    
    filename = os.path.basename(nzbfile.filename)
    
    logging.info('[%s] Adding %s', __NAME__, filename)
    try:
        NZBQ.add(NzbObject(filename, nzbfile.value, repair, 
                           unpack, delete))
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
        
@synchronized_CV
@synchronized
def add_nzo(nzo, position = -1):
    try:
        NZBQ.add(nzo, position)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
        
@synchronized_CV
@synchronized
def insert_future_nzo(future_nzo, nzo):
    try:
        NZBQ.insert_future(future_nzo, nzo)
    except NameError:
        logging.exception("[%s] Error accessing NZBQ?", __NAME__)
        
@synchronized_CV
def pause_downloader():
    try:
        DOWNLOADER.pause()
    except NameError:
        logging.exception("[%s] Error accessing DOWNLOADER?", __NAME__)
        
@synchronized_CV
def resume_downloader():
    try:
        DOWNLOADER.resume()
    except NameError:
        logging.exception("[%s] Error accessing DOWNLOADER?", __NAME__)
         
################################################################################
## Unsynchronized methods                                                     ##
################################################################################
def postprocess_nzo(nzo):
    try:
        POSTPROCESSOR.process(nzo)
    except NameError:
        logging.exception("[%s] Error accessing POSTPROCESSOR?", __NAME__)
        
def assemble_nzf(nzf):
    try:
        ASSEMBLER.process(nzf)
    except NameError:
        logging.exception("[%s] Error accessing ASSEMBLER?", __NAME__)
        
def disconnect():
    try:
        DOWNLOADER.disconnect()
    except NameError:
        logging.exception("[%s] Error accessing DOWNLOADER?", __NAME__)
        
def paused():
    try:
        return DOWNLOADER.paused
    except NameError:
        logging.exception("[%s] Error accessing DOWNLOADER?", __NAME__)
            
def bps():
    try:
        return DOWNLOADER.meter.bps
    except NameError:
        logging.exception("[%s] Error accessing DOWNLOADER?", __NAME__)
        
################################################################################
# Helpers                                                                      #
################################################################################
def pp_to_opts(pp):
    repair, unpack, delete = (False, False, False)
    if pp > 0:
        repair = True
        if pp > 1:
            unpack = True
            if pp > 2:
                delete = True
    return (repair, unpack, delete)
