# Copyright 2005 Gregor Kaufmann <tdn@gmx.at>
# Licensed under the Academic Free License 2.1
""" 
sabnzbd.assembler - threaded assembly/decoding of files
"""
#------------------------------------------------------------------------------

__NAME__ = "sabnzbd.assembler"

import os
import re
import Queue
import binascii
import logging
import sabnzbd
import cPickle
import shutil

from sabnzbd.newsunpack import unpack_magic, par2_repair
from threading import Thread
from time import sleep

#------------------------------------------------------------------------------

## sabnzbd.add_nzo
## sabnzbd.cleanup_nzo
class PostProcessor(Thread):
    def __init__ (self, download_dir, complete_dir, queue = None):
        Thread.__init__(self)
        
        self.download_dir = download_dir
        self.complete_dir = complete_dir
        self.queue = queue
        
        if not self.queue:
            self.queue = Queue.Queue()
        
    def process(self, nzo):
        self.queue.put(nzo)
        
    def stop(self):
        self.queue.put(None)
        
    def run(self):
        while 1:
            nzo = self.queue.get()
            if not nzo:
                break
                
            try:
                rep, unp, dele = nzo.get_repair_opts()
                
                repairsets = nzo.get_repairsets()
                dirname = nzo.get_dirname()
                filename = nzo.get_filename()
                groupname = nzo.get_group()
                
                if nzo.create_group_folder:
                    workdir = os.path.join(self.download_dir, groupname, dirname)
                else:
                    workdir = os.path.join(self.download_dir, dirname)
                
                                                   
                logging.info('[%s] Starting PostProcessing on %s' + \
                             ' => Repair:%s, Unpack:%s, Delete:%s',
                             __NAME__, filename, rep, unp, dele)
                
                ## Run Stage 1: Repair
                if rep:
                    logging.info('[%s] Par2 check starting on %s', __NAME__, filename)
                    readd = False
                    if not repairsets:
                        logging.info("[%s] No par2 sets for %s", __NAME__, filename)
                        nzo.set_unpackstr('=> No par2 sets', '[PAR-INFO]', 1)
                        
                    for _set in repairsets:
                        logging.info("[%s] Running repair on set %s", __NAME__, _set)
                        parfile_nzf = nzo.get_partable_by_setname(_set)
                        need_readd = par2_repair(parfile_nzf, nzo, workdir, _set)
                        if need_readd:
                            readd = True
                            
                    if readd:
                        logging.info('[%s] Readded %s to queue', __NAME__, filename)
                        sabnzbd.add_nzo(nzo, 0)
                        
                        ## Break out
                        continue
                        
                    logging.info('[%s] Par2 check finished on %s', __NAME__, filename)
                    
                workdir_complete = None
                
                
                if self.complete_dir:
                    dirname = "__UNPACK_IN_PROGRESS__%s" % nzo.get_dirname()
                    workdir_complete = get_dirpath(nzo.get_group(), dirname, 
                                                   nzo, self.complete_dir)
                    
                ## Run Stage 2: Unpack
                error = False
                if unp:
                    error = unpack_magic(nzo, workdir, workdir_complete, dele, (), (), ())
                        
                ## Run Stage 3: Cleanup
                # Move remaining files to workdir_complete
                if workdir_complete:
                    try:
                        for root, dirs, files in os.walk(workdir):
                            for _file in files:
                                path = os.path.join(root, _file)
                                new_path = path.replace(workdir, workdir_complete)
                                shutil.move(path, new_path)
                        os.rmdir(workdir)
                        
                        workdir_final = workdir_complete.replace("__UNPACK_IN_PROGRESS__", "")
                        
                        os.rename(workdir_complete, workdir_final)
                        
                        workdir = workdir_final
                    except:
                        logging.exception("[%s] Moving files/dirs in %s to %s failed",
                                          __NAME__, workdir, workdir_complete)
                                          
                if sabnzbd.CLEANUP_LIST:
                    try:
                        files = os.listdir(workdir)
                    except:
                        files = ()
                        
                    for _file in files:
                        root, ext = os.path.splitext(_file)
                        
                        if ext in sabnzbd.CLEANUP_LIST:
                            path = os.path.join(workdir, _file)
                            try:
                                logging.info("[%s] Removing unwanted file %s",
                                             __NAME__, path)
                                os.remove(path)
                            except:
                                logging.exception("[%s] Removing %s failed",
                                                  __NAME__, path)
                                                   
                logging.info('[%s] %s done! Cleaning up.', __NAME__, filename)
                sabnzbd.cleanup_nzo(nzo)
                
            except:
                logging.exception("[%s] Postprocessing of %s failed.", __NAME__,
                                  nzo.get_filename())
#------------------------------------------------------------------------------

## sabnzbd.pause_downloader
class Assembler(Thread):
    def __init__ (self, download_dir, cache_dir_articles, queue = None):
        Thread.__init__(self)
        
        self.download_dir = download_dir
        self.cache_dir_articles = cache_dir_articles
        self.queue = queue
        
        if not self.queue:
            self.queue = Queue.Queue()
            
    def stop(self):
        self.process(None)
        
    def process(self, nzf):
        self.queue.put(nzf)
    
    def assemble(self, nzf, path, ram_cached):
        fout = open(path, 'ab')
        
        _type = nzf.get_type()
        datatable = nzf.get_datatable()
        decodetable = nzf.get_decodetable()
        
        for articlenum in decodetable:
            sleep(0.01)
            article = decodetable[articlenum]
            data = None
            if ram_cached:
                data = datatable[article]
                datatable[article] = None
            else:
                path = os.path.join(self.cache_dir_articles, article.article)
                if os.path.exists(path):
                    ftmp = open(path, 'rb')
                    if _type == 'yenc':
                        data = ftmp.read()
                    elif _type == 'uu':
                        data = cPickle.load(ftmp)
                    ftmp.close()
                    try:
                        os.remove(path)
                    except OSError:
                        logging.exception("[%s] Removing %s failed.", __NAME__, 
                                          path)
                                          
            if not data:
                logging.warning('[%s] %s missing', __NAME__, article.article)
            else:
                # yenc data already decoded, flush it out
                if _type == 'yenc':
                    fout.write(data)
                # need to decode uu data now
                elif _type == 'uu':
                    chunks = []
                    for line in data:
                        try:
                            tmpdata = binascii.a2b_uu(line)
                            chunks.append(tmpdata)
                        except binascii.Error, msg:
                            ## Workaround for broken uuencoders by 
                            ##/Fredrik Lundh
                            nbytes = (((ord(line[0])-32) & 63) * 4 + 5) / 3
                            try:
                                tmpdata = binascii.a2b_uu(line[:nbytes])
                                chunks.append(tmpdata)
                            except binascii.Error, msg:
                                logging.info('[%s] Decode failed in part %s: %s',
                                             __NAME__, article.article, msg)
                    fout.write(''.join(chunks))
        fout.flush()
        fout.close()
        
    def run(self):
        while 1:
            nzo_nzf_tuple = self.queue.get()
            if not nzo_nzf_tuple:
                logging.info("[%s] Shutting down", __NAME__)
                break
            
            nzo, nzf = nzo_nzf_tuple
            
            if nzf:
                try:
                    dirname = nzo.get_dirname()
                    groupname = nzo.get_group()
                    filename = nzf.get_filename()
                    _type = nzf.get_type()
                    
                    # Get filepath, setup dirs on the way
                    filepath = get_filepath(self.download_dir, groupname, dirname, 
                                            filename, nzo)
                    if filepath and not os.path.exists(filepath):
                        logging.info('[%s] Decoding %s %s', __NAME__, filepath, 
                                     _type)
                        try:
                            self.assemble(nzf, filepath, nzo.ram_cached)
                        except IOError, (errno, strerror):
                            # 28 == disk full => pause downloader
                            if errno == 28:
                                sabnzbd.pause_downloader()
                                logging.warning('[%s] Disk full! Forcing Pause', 
                                                __NAME__)
                            else:
                                logging.exception('[%s] Disk exception!' + \
                                                  ' Please report this.', __NAME__)
                                fixed_filename = fix_filename(filename)
                                if fixed_filename != filename:
                                    logging.info('[%s] Retrying %s with new' + \
                                                 ' filename %s', 
                                                 __NAME__, filename, fixed_filename)
                                    try:
                                        filepath = get_filepath(self.download_dir, 
                                                                groupname, dirname,
                                                                fixed_filename, nzo)
                                        self.assemble(nzf, filepath, nzo.ram_cached)
                                    except Exception:
                                        pass
                                        
                except:
                    logging.exception("[%s] Assembly of %s failed.", __NAME__, nzf)
                
            else:
                sabnzbd.postprocess_nzo(nzo)
                
#------------------------------------------------------------------------------

def get_filepath(download_dir, groupname, dirname, filename, nzo):
    dirpath = None
    filepath = None
    
    if nzo.create_group_folder:
        logging.debug("[%s] Using group folders", __NAME__)
        groupdirpath = create_dir(os.path.join(download_dir, groupname))
        
        if groupdirpath:
            dirpath = create_dir(os.path.join(groupdirpath, dirname))
    else:
        logging.debug("[%s] Not using group folders", __NAME__)
        dirpath = create_dir(os.path.join(download_dir, dirname))
        
    if dirpath:
        filepath = os.path.join(dirpath, filename)
        
    return filepath
    
def get_dirpath(groupname, dirname, nzo, complete_dir):
    dirpath = None
    
    if nzo.create_group_folder:
        logging.debug("[%s] Using group folders", __NAME__)
        groupdirpath = create_dir(os.path.join(complete_dir, groupname))
        if groupdirpath:
            dirpath = create_dir(os.path.join(groupdirpath, dirname))
    else:
        logging.debug("[%s] Not using group folders", __NAME__)
        dirpath = create_dir(os.path.join(complete_dir, dirname))
        
    return dirpath
    
def create_dir(dirpath):
    if not os.path.exists(dirpath):
        logging.info('[%s] Creating directory: %s', __NAME__, dirpath)
        try:
            os.mkdir(dirpath, 0755)
        except OSError:
            logging.exception('[%s] Creating directory %s failed', __NAME__,
                              dirpath)
            return None
            
    return dirpath
    
PROPER_FILENAME_MATCHER = re.compile(r"[a-zA-Z0-9\-_\.+]")
def fix_filename(filename):
    cst = []
    for i in xrange(len(filename)):
        if PROPER_FILENAME_MATCHER.search(filename[i]):
            cst.append(filename[i])
        else:
            cst.append("_")
    filename = ''.join(cst)
    return filename
