# Copyright 2005 Gregor Kaufmann <tdn@gmx.at>
# Licensed under the Academic Free License 2.1
"""
sabnzbd.downloader - download engine
"""

__NAME__ = 'sabnzbd.downloader'

import time
import select
import logging
import sabnzbd

from threading import Thread
from sabnzbd.newswrapper import NewsWrapper
from sabnzbd.nzbqueue import CrcError

#------------------------------------------------------------------------------

class Server:
    def __init__(self, host, port, threads, fillserver, username = None, 
                 password = None):
        self.host = host
        self.port = port
        self.threads = threads
        self.fillserver = fillserver
        
        self.username = username
        self.password = password
        
        self.busy_threads = []
        self.idle_threads = []
        
        for i in range(threads):
            self.idle_threads.append(NewsWrapper(self, i+1))
            
#------------------------------------------------------------------------------

class BPSMeter:
    def __init__(self):
        t = time.time()
        
        self.start_time = t
        self.log_time = t
        self.last_update = t
        self.bps = 0.0
        self.bytes_total = 0
        
    def update(self, bytes_recvd):
        self.bytes_total += bytes_recvd
        t = time.time()
        try:
            self.bps = (self.bps * (self.last_update - self.start_time)
                        + bytes_recvd) / (t - self.start_time)
        except:
            self.bps = 0.0
            
        self.last_update = t
        
        check_time = t - 20.0
        
        if self.start_time < check_time:
            self.start_time = check_time
        
        if self.log_time < check_time:
            logging.info("[%s] bps: %s", __NAME__, self.bps)
            self.log_time = t
            
        if self.bps < 0.0:
            logging.debug("[%s] bps < 0 -> reset", __NAME__) 
            self.reset()
        
    def reset(self):
        self.__init__()
        
#------------------------------------------------------------------------------

class Downloader(Thread):
    def __init__(self, servers, bandwith_limit, paused = False):
        Thread.__init__(self)
        
        self.bandwith_limit = bandwith_limit
        
        self.paused = paused
        
        self.shutdown = False
        
        self.active = False
        
        self.force_disconnect = False
        
        self.read_fds = {}
        self.write_fds = {}
        
        self.servers = []
        
        for server in servers:
            host = servers[server]['host']
            port = int(servers[server]['port'])
            threads = int(servers[server]['connections'])
            fillserver = bool(int(servers[server]['fillserver']))
            username = servers[server]['username']
            password = servers[server]['password']
            self.servers.append(Server(host, port, threads, fillserver, 
                                       username, password))
                                       
        self.meter = BPSMeter()
        
    def stop(self):
        self.shutdown = True
        
    def resume(self):
        self.paused = False
        
    def pause(self):
        self.paused = True
        
    def bps(self):
        return self.meter.bps
        
    def disconnect(self):
        self.force_disconnect = True
        
    def run(self):
        self.active = True
        
        while 1:
            for server in self.servers:
                for nw in server.busy_threads[:]:
                    if nw.timeout and time.time() > nw.timeout:
                        self.__reset_nw(nw, "timed out")
                        
                if not server.idle_threads or self.paused or self.shutdown:
                    continue
                    
                if not sabnzbd.has_articles_for(server.host):
                    continue
                    
                for nw in server.idle_threads[:]:
                    if nw.timeout:
                        if time.time() < nw.timeout:
                            continue
                        else:
                            nw.timeout = None
                            
                    article = sabnzbd.get_article(server.host, server.fillserver)
                    
                    if not article:
                        break
                        
                    else:
                        server.idle_threads.remove(nw)
                        server.busy_threads.append(nw)
                        
                        nw.article = article
                        if nw.connected:
                            self.__request_article(nw)
                            
                        else:
                            try:
                                logging.info("[%s] %s@%s:%s: Initiating connection",
                                                  __NAME__, nw.thrdnum, server.host,
                                                  server.port) 
                                nw.init_connect()
                                self.write_fds[nw.nntp.sock.fileno()] = nw
                            except:
                                logging.exception("[%s] Failed to initialize %s@%s:%s",
                                                  __NAME__, nw.thrdnum, server.host,
                                                  server.port)
                                self.__reset_nw(nw, "failed to initialize")
                                
            # Exit-point
            if self.shutdown:
                empty = True
                for server in self.servers:
                    if server.busy_threads:
                        empty = False
                        break
                        
                if empty:
                    for server in self.servers:
                        for nw in server.idle_threads:
                            nw.hard_reset()
                            
                    self.active = False
                    logging.info("[%s] Shutting down", __NAME__)
                    break
                    
            if self.force_disconnect:
                for server in self.servers:
                    for nw in server.busy_threads[:]:
                        self.__reset_nw(nw, "forcing disconnect")
                        
                    for nw in server.idle_threads[:]:
                        self.__reset_nw(nw, "forcing disconnect")
                        
                self.force_disconnect = False
                
            # => Select
            readkeys = self.read_fds.keys()
            writekeys = self.write_fds.keys()
            
            if readkeys or writekeys:
                read, write, error = select.select(readkeys, writekeys, (), 1.0)
                
            else:
                read, write, error = ([], [], [])
                
                self.meter.reset()
                
                time.sleep(1.0)
                
                sabnzbd.CV.acquire()
                while (not sabnzbd.has_articles() or self.paused) and not \
                       self.shutdown:
                    sabnzbd.CV.wait()
                sabnzbd.CV.release()
                
                self.force_disconnect = False
                
            for selected in write:
                nw = self.write_fds[selected]
                
                fileno = nw.nntp.sock.fileno()
                
                if fileno not in self.read_fds:
                    self.read_fds[fileno] = nw
                    
                if fileno in self.write_fds:
                    self.write_fds.pop(fileno)
                    
            if not read:
                self.meter.update(0)
                continue
                
            for selected in read:
                nw = self.read_fds[selected]
                article = nw.article
                server = nw.server
                
                if article:
                    nzo = article.nzf.nzo
                    
                try:
                    if self.bandwith_limit:
                        time.sleep(self.bandwith_limit)
                    bytes, done = nw.recv_chunk()
                except:
                    bytes, done = (0, False)
                    
                if bytes < 1:
                    self.__reset_nw(nw, "server closed connection")
                    continue
                    
                else:
                    self.meter.update(bytes)
                    
                    if nzo:
                        nzo.bytes_downloaded += bytes
                        
                #Check for 423 & 430 (Article not found)
                if len(nw.lines) == 1:
                    if not nw.connected:
                        try:
                            nw.finish_connect()
                            logging.debug("[%s] %s@%s:%s last message -> %s", 
                                         __NAME__, nw.thrdnum, nw.server.host,
                                         nw.server.port, nw.lines[0])
                            nw.lines = []
                        except:
                            logging.exception("[%s] Connecting %s@%s:%s failed", 
                                              __NAME__, nw.thrdnum, nw.server.host,
                                              nw.server.port)
                            self.__reset_nw(nw, "connecting failed")
                            
                        if nw.connected:
                            logging.info("[%s] Connecting %s@%s:%s finished", 
                                         __NAME__, nw.thrdnum, nw.server.host,
                                         nw.server.port)
                            self.__request_article(nw)
                            
                        continue
                        
                    if nw.lines[0][:3] in ('423', '430'):
                        self.__search_new_server(article, nw)
                        continue
                        
                if done:
                    try:
                        sabnzbd.register_article(article, nw.lines)
                        logging.info('[%s] Thread %s@%s:%s: %s done',
                                     __NAME__, nw.thrdnum, server.host, 
                                     server.port, article.article)
                    except IOError, e:
                        self.paused = True
                        
                        logging.exception('[%s] Thread %s@%s:%s: IOError',
                                          __NAME__, nw.thrdnum, server.host,
                                          server.port)
                        logging.info('[%s] Thread %s@%s:%s: Forcing pause',
                                     __NAME__, nw.thrdnum, server.host,
                                     server.port)
                                     
                        self.__reset_nw(nw, "forcing disconnect")
                        continue
                        
                    except CrcError, e:
                        logging.info('[%s] Thread %s@%s:%s: %s crc ' + \
                                     'error (need: %s, got: %s)',
                                     __NAME__, nw.thrdnum, server.host, 
                                     server.port, article.article,
                                     e.needcrc, e.gotcrc)
                                          
                        if sabnzbd.FAIL_ON_CRC:
                            self.__search_new_server(article, nw)
                            continue
                            
                        else:
                            logging.info('[%s] Thread %s@%s:%s: %s, discarding',
                                         __NAME__, nw.thrdnum, server.host, 
                                         server.port, article.article)
                            sabnzbd.register_article(nw.article)
                            
                    nw.soft_reset()
                    server.busy_threads.remove(nw)
                    server.idle_threads.append(nw)
                    
    def __reset_nw(self, nw, errormsg):
        server = nw.server
        article = nw.article
        fileno = None
        
        if nw.nntp:
            fileno = nw.nntp.sock.fileno()
            
        logging.warning('[%s] Thread %s@%s:%s: ' + errormsg,
                         __NAME__, nw.thrdnum, server.host, server.port)
                         
        if nw in server.busy_threads:
            server.busy_threads.remove(nw)
        if nw not in server.idle_threads:
            server.idle_threads.append(nw)
            
        if fileno and fileno in self.write_fds:
            self.write_fds.pop(fileno)
        if fileno and fileno in self.read_fds:
            self.read_fds.pop(fileno)
            
        # Remove this server from try_list
        if article:
            article.try_list.remove(nw.server.host)
            
            article.fetched = False
            
            # Cleanup try lists
            self.__cleanup_try_lists(article)
            
        nw.hard_reset()
        
    def __request_article(self, nw):
        try:
            logging.info('[%s] Thread %s@%s:%s: fetching %s',
                         __NAME__, nw.thrdnum, nw.server.host, 
                         nw.server.port, nw.article.article)
                         
            fileno = nw.nntp.sock.fileno()
            
            nw.body(nw.article)
            
            if fileno not in self.read_fds:
                self.read_fds[fileno] = nw
        except:
            self.__reset_nw(nw, "server closed connection")
            
    def __cleanup_try_lists(self, article):
        nzf = article.nzf
        nzo = nzf.nzo
        
        new_server_found = False
        fill_server_found = False
        
        for server in self.servers:
            host = server.host
            if host not in article.try_list:
                if server.fillserver:
                    fill_server_found = True
                else:
                    new_server_found = True
                    
                ## Allow all servers to iterate over each nzo/nzf again ##
                nzf.remove_from_try_list(host)
                nzo.remove_from_try_list(host)
                sabnzbd.remove_from_try_list(host)
                
        # Only found one (or more) fill server(s)
        if not new_server_found and fill_server_found:
            article.allow_fill_server = True
            new_server_found = True
            
        return new_server_found
        
    def __search_new_server(self, article, nw):
        logging.warning('[%s] Thread %s@%s:%s: Article ' + \
                        '%s missing, searching new server...',
                        __NAME__, nw.thrdnum, nw.server.host, 
                        nw.server.port, article.article)
                        
        nw.server.busy_threads.remove(nw)
        nw.server.idle_threads.append(nw)
        nw.soft_reset()
        
        new_server_found = self.__cleanup_try_lists(article)
        
        if new_server_found:
            article.fetched = False
            
            logging.warning('[%s] Thread %s@%s:%s: Article ' + \
                            '%s => found at least one untested server',
                            __NAME__, nw.thrdnum, nw.server.host,
                            nw.server.port, article.article)
                            
        else:
            logging.warning('[%s] Thread %s@%s:%s: Article ' + \
                            '%s missing from all servers, discarding',
                            __NAME__, nw.thrdnum, nw.server.host,
                            nw.server.port, article.article)
                            
            sabnzbd.register_article(article)
