"""
Copyright (c) 2005, Sylvain Hellegouarch
Copyright (c) 2005, Tiago Cogumbreiro <cogumbreiro@users.sf.net>

All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, 
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution.
    * Neither the name of Sylvain Hellegouarch nor the names of his contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

#
# This is a simple filter for Digest and Basic Authorization schemes
# as described in RFC 2617
# http://www.ietf.org/rfc/rfc2617.txt
#

import time
import os.path
from stat import ST_MTIME
from os import stat

import cherrypy
from cherrypy.lib.filter.basefilter import BaseFilter
import httpauth

class HttpAuthFilter(BaseFilter):
    """Class that must be called in your filters list.
    It requires a path to the cppassword file (which can be modified after the server has been started)
    A realm value (see RFC 2617)
    A private key to compute the nonce field
    An unauthorizedPath which is a path (as a string) to the exposed object to call in case the authorization fails
    A defaultPath in case something wrong happens. By default '/default'
    """
    def __init__(self, cppasswordPath, realm, privateKey, unauthorizedPath, defaultPath='/default'):
        #self.cppasswordPath = cppasswordPath
        un, pw = cppasswordPath

        self.cppasswordLastModified = None
        self.users = {}
        self.users[un] = pw
        self.realm = realm
        self.privateKey = privateKey
        self.unauthorizedPath = unauthorizedPath
        self.defaultPath = defaultPath

    def _loadAuthorizedUsers(self):
        # We don't to reload the file for each request. Let's reload it only when it is modified
        #if not self.cppasswordLastModified or self.cppasswordLastModified != stat(self.cppasswordPath)[ST_MTIME]:
        #    self.cppasswordLastModified = stat(self.cppasswordPath)[ST_MTIME]
        #    cppassword = file(self.cppasswordPath, 'rb')
        #    for line in cppassword:
        #        username, password = line.split(':')
        #        self.users[username] = password.rstrip()
        #    cppassword.close()
        return self.users

    def beforeRequestBody(self):
        # Make sure it will not be authorized whatever happens in between
        cherrypy.threadData.isAuthorized = False
        cherrypy.threadData.hasBeenProcessed = False
        # Check if we have an authorization request header
        if cherrypy.request.headerMap.has_key('Authorization'):
            # Parse the header
            ah = httpauth.parseAuthorization(cherrypy.request.headerMap['Authorization'])
            # If one required field is missing, default(). Maybe we could look a message as well
            if ah is None:
                cherrypy.request.objectPath = self.defaultPath
                cherrypy.response.status = '400 Bad Request'
            else:
                # Is the user OK?
                password = self._loadAuthorizedUsers().get(ah["username"], None)
                cherrypy.threadData.hasBeenProcessed = True
                 
                if httpauth.checkResponse(ah, password, method=cherrypy.request.method):
                    cherrypy.threadData.isAuthorized = True
                    cherrypy.response.status = '200 OK' # just in case
                else:
                    cherrypy.request.objectPath = self.unauthorizedPath
        else:
            cherrypy.request.objectPath = self.unauthorizedPath

    def onEndResource(self):
        # Be sure we ask the user to tell us who he/she is
        if cherrypy.threadData.hasBeenProcessed and cherrypy.request.objectPath is self.unauthorizedPath:
            cherrypy.response.status = '401 Unauthorized'
        elif cherrypy.threadData.isAuthorized is False:
            cherrypy.response.status = '401 Unauthorized'
            if cherrypy.response.headers:
                cherrypy.response.headers.append(('WWW-Authenticate', httpauth.digestAuth (self.realm)))
                cherrypy.response.headers.append(('WWW-Authenticate', httpauth.basicAuth (self.realm)))
            
if __name__ == '__main__':
    class Root:
        def index(self):
            return "hey how you doing?"
        index.exposed = True
    
    class Admin:
        _cpFilterList = [ HttpAuthFilter(cppasswordPath='.bloopwd',
                                         realm='localhost',
                                         privateKey='duh!',
                                         unauthorizedPath='/admin/unauthorized',
                                         defaultPath='/admin/default') ]
        def index(self):
            '''Only available once the user is authenticated'''
            return "Hello there!"
        index.exposed = True

        def default(self):
            '''In case the user agent tried to trick us or is faulty'''
            return "Bad request"
        default.exposed = True

        def unauthorized(self):
            '''Must be declared at the same level where you definedt he filter'''
            return "You are not authorized to access that page"
        unauthorized.exposed = True

    class Son:
        def index(self):
            '''Protected by the filter above'''
            return "Hello son!"
        index.exposed = True

    cherrypy.root = Root()
    cherrypy.root.admin = Admin()
    cherrypy.root.admin.son = Son()
    cherrypy.server.start()
