# Copyright 2005 Gregor Kaufmann <tdn@gmx.at>
# Licensed under the Academic Free License 2.1
"""
sabnzbd.interface - webinterface
"""

__NAME__ = "sabnzbd.interface"

import os
import datetime
import time
import cherrypy
import sabnzbd
import logging
import random
import shutil
import listquote

import httpauthfilter_sabnzbd

from configobj import ConfigObj
from Cheetah.Template import Template

#------------------------------------------------------------------------------

try:
    os.statvfs
    import statvfs
    # posix diskfree
    def diskfree(_dir):
        try:
            s = os.statvfs(_dir)
            return (s[statvfs.F_BFREE] * s[statvfs.F_FRSIZE]) / 1073741824.0
        except OSError:
            return 0.0
except AttributeError:
    try:
        import win32api
    except ImportError:
        pass
    # windows diskfree
    def diskfree(_dir):
        try:
            secp, byteper, freecl, noclu = win32api.GetDiskFreeSpace(_dir)
            return (secp * byteper * freecl) / 1073741824.0
        except:
            return 0.0
            
#------------------------------------------------------------------------------

class BlahPage:
    def index(self):
        return ""
    index.exposed = True
    
class MainPage:
    def __init__(self, username, password, web_dir):
        self.__root = '/sabnzbd'
        
        self.__web_dir = web_dir
        
        self._cpfilterList = []
        if username and password:
            pw = (username, password)
            self._cpFilterList = [ httpauthfilter_sabnzbd.HttpAuthFilter(pw,
                                   realm='SABnzbd',
                                   privateKey=genKey(),
                                   unauthorizedPath='/sabnzbd/walk',
                                   defaultPath='/') ]
                                   
    def index(self):
        info, postinfo, bytespersec = build_header()
        
        if sabnzbd.USERNAME_NEWZBIN and sabnzbd.PASSWORD_NEWZBIN:
            info['newzbinDetails'] = True
            
        template = Template(file=os.path.join(self.__web_dir, 'main.tmpl'),
                            searchList=[info],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True
    
    def walk(self, *args):
        return "Auth failed."
    walk.exposed = True
    
    def addID(self, id = None, pp = 0):
        if id:
            id = id.strip()
            
        if id and id.isdigit() and pp.isdigit():
            sabnzbd.add_msgid(int(id), int(pp))
        raise cherrypy.HTTPRedirect(self.__root)
    addID.exposed = True
    
    def addURL(self, url = None, pp = 0):
        if url and pp.isdigit():
            sabnzbd.add_url(url, int(pp))
        raise cherrypy.HTTPRedirect(self.__root)
    addURL.exposed = True
    
    def addFile(self, nzbfile, pp = 0):
        if pp.isdigit():
            sabnzbd.add_nzbfile(nzbfile, int(pp))
        raise cherrypy.HTTPRedirect(self.__root)
    addFile.exposed = True
    
    def shutdown(self):
        yield "Initiating shutdown..."
        sabnzbd.halt()
        cherrypy.server.stop()
        yield "<br>SABnzbd-%s shutdown finished" % sabnzbd.__version__
        raise KeyboardInterrupt()
    shutdown.exposed = True
    
    def pause(self):
        sabnzbd.pause_downloader()
        raise cherrypy.HTTPRedirect(self.__root)
    pause.exposed = True
    
    def resume(self):
        sabnzbd.resume_downloader()
        raise cherrypy.HTTPRedirect(self.__root)
    resume.exposed = True
    
def genKey():
    key = ''
    for i in xrange(random.randint(200, 400)):
        key += '%c' % random.randint(0, 255)
    return key
    
#------------------------------------------------------------------------------

class QueuePage:
    def __init__(self, web_dir):
        self.__root = '/sabnzbd/queue'
        
        self.__web_dir = web_dir
        
        self.__verbose = False
        
    def index(self):
        info, postinfo, bytespersec = build_header()
        
        info['noofslots'] = len(postinfo)
        datestart = datetime.datetime.now()
        
        n = 0
        slotinfo = []
        for post in postinfo:
            post_id, nfo = post
            repair, unpack, delete, filename, unpackstr, bytesleft, bytes, \
                  average_date, finished_files, active_files, queued_files = nfo
            slot = {'index':n, 'post_id':str(post_id)}
            n += 1
            r = 'No'
            u = 'No'
            d = 'No'
            if repair:
                r = 'Yes'
                if unpack:
                    u = 'Yes'
                    if delete:
                        d = 'Yes'
            slot['repair'] = r
            slot['unpack'] = u
            slot['delete'] = d
            slot['filename'] = filename
            slot['mbleft'] = "%.2f" % (bytesleft/1048576.0)
            slot['mb'] = "%.2f" % (bytes/1048576.0)
            
            try:
                datestart = datestart + datetime.timedelta(seconds=bytesleft / bytespersec)
                slot['eta'] = '%s' % datestart.ctime()
            except:
                datestart = datetime.datetime.now()
                slot['eta'] = 'unknown'
                
            slot['avg_age'] = calc_age(average_date)
                
            finished = []
            active = []
            queued = []
            if self.__verbose:
                
                date_combined = 0
                num_dates = 0
                
                for tup in finished_files:
                    bytes_left, bytes, fn, date = tup
                    if isinstance(fn, unicode): 
                        fn = fn.encode('utf-8')
                        
                    age = calc_age(date)
                    
                    line = {'filename':str(fn), 
                            'mbleft':"%.2f" % (bytes_left/1048576.0), 
                            'mb':"%.2f" % (bytes/1048576.0),
                            'age':age}
                    finished.append(line)
                    
                for tup in active_files:
                    bytes_left, bytes, fn, date, nzf_id = tup
                    if isinstance(fn, unicode): 
                        fn = fn.encode('utf-8')
                        
                    age = calc_age(date)
                    
                    line = {'filename':str(fn), 
                            'mbleft':"%.2f" % (bytes_left/1048576.0), 
                            'mb':"%.2f" % (bytes/1048576.0), 
                            'nzf_id':nzf_id,
                            'age':age}
                    active.append(line)
                    
                for tup in queued_files:
                    _set, bytes_left, bytes, fn, date = tup
                    if isinstance(fn, unicode): 
                        fn = fn.encode('utf-8')
                    
                    age = calc_age(date)
                    
                    line = {'filename':str(fn), 'set':_set, 
                            'mbleft':"%.2f" % (bytes_left/1048576.0), 
                            'mb':"%.2f" % (bytes/1048576.0),
                            'age':age}
                    queued.append(line)
                    
            slot['finished'] = finished
            slot['active'] = active
            slot['queued'] = queued
            
            slotinfo.append(slot)
            
        if slotinfo:
            info['slotinfo'] = slotinfo
            
        template = Template(file=os.path.join(self.__web_dir, 'queue.tmpl'),
                            searchList=[info],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True  
    
    def delete(self, uid = None):
        if uid:
            sabnzbd.remove_nzo(uid, False)
        raise cherrypy.HTTPRedirect(self.__root)
    delete.exposed = True
    
    def removeNzf(self, nzo_id = None, nzf_id = None):
        if nzo_id and nzf_id:
            sabnzbd.remove_nzf(nzo_id, nzf_id)
        raise cherrypy.HTTPRedirect(self.__root)
    removeNzf.exposed = True
    
    def tog_verbose(self):
        self.__verbose = not self.__verbose
        raise cherrypy.HTTPRedirect(self.__root)
    tog_verbose.exposed = True
    
    def switch(self, uid1 = None, uid2 = None):
        if uid1 and uid2:
            sabnzbd.switch(uid1, uid2)
        raise cherrypy.HTTPRedirect(self.__root)
    switch.exposed = True
    
    def shutdown(self):
        yield "Initiating shutdown..."
        sabnzbd.halt()
        cherrypy.server.stop()
        yield "<br>SABnzbd-%s shutdown finished" % sabnzbd.__version__
        raise KeyboardInterrupt()
    shutdown.exposed = True
    
    def pause(self):
        sabnzbd.pause_downloader()
        raise cherrypy.HTTPRedirect(self.__root)
    pause.exposed = True
    
    def resume(self):
        sabnzbd.resume_downloader()
        raise cherrypy.HTTPRedirect(self.__root)
    resume.exposed = True
    
    def sort_by_avg_age(self):
        sabnzbd.sort_by_avg_age()
        raise cherrypy.HTTPRedirect(self.__root)
    sort_by_avg_age.exposed = True
    
class HistoryPage:
    def __init__(self, web_dir):
        self.__root = '/sabnzbd/history'
        
        self.__web_dir = web_dir
        
        self.__verbose = True
        
    def index(self):
        history, postinfo, bytespersec = build_header()
            
        history_items, total_bytes = sabnzbd.history_info()
        
        history['total_bytes'] = "%.2f" % (total_bytes / 1073741824.0)
        
        stagenames = {1:"Par2", 2:"Unrar", 3:"Unzip", 4:"Filejoin"}
        
        items = []
        while history_items:
            added = max(history_items.keys())
            
            history_item_list = history_items.pop(added)
            
            for history_item in history_item_list:
                filename, unpackstrht, loaded = history_item
                stages = []
                item = {'added':time.ctime(added), 'filename':filename, 'loaded':loaded, 'stages':stages}
                if self.__verbose:
                    for stage in unpackstrht:
                        stageLine = {'name':stagenames[stage]}
                        actions = []
                        for action in unpackstrht[stage]:
                            actionLine = {'name':action, 'value':unpackstrht[stage][action]}
                            actions.append(actionLine)
                        stageLine['actions'] = actions
                        stages.append(stageLine)
                item['stages'] = stages
                items.append(item)
        history['lines'] = items
        
        template = Template(file=os.path.join(self.__web_dir, 'history.tmpl'),
                            searchList=[history],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True
    
    def purge(self):
        sabnzbd.purge_history()
        raise cherrypy.HTTPRedirect(self.__root)
    purge.exposed = True
    
    def reset(self):
        sabnzbd.reset_byte_counter()
        raise cherrypy.HTTPRedirect(self.__root)
    reset.exposed = True
    
    def tog_verbose(self):
        self.__verbose = not self.__verbose
        raise cherrypy.HTTPRedirect(self.__root)
    tog_verbose.exposed = True
    
#------------------------------------------------------------------------------

class ConfigPage:
    def __init__(self, web_dir):
        self.__root = '/sabnzbd/config'
        
        self.__web_dir = web_dir
        
    def index(self):
        config, postinfo, bytespersec = build_header()
        
        if type(sabnzbd.CFG.filename) is type(''):
            fn = sabnzbd.CFG.filename
        else:
            fn = sabnzbd.CFG.filename.name
        
        config['configfn'] = fn
            
        template = Template(file=os.path.join(self.__web_dir, 'config.tmpl'),
                            searchList=[config],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True
    
    def restart(self):
        sabnzbd.halt()
        init_ok = sabnzbd.initialize()
        if init_ok:
            sabnzbd.start()
            raise cherrypy.HTTPRedirect(self.__root)
        else:
            return "SABnzbd restart failed! See logfile(s)."
    restart.exposed = True
    
#------------------------------------------------------------------------------

class ConfigDirectories:
    def __init__(self, web_dir):
        self.__root = '/sabnzbd/config/directories'
        
        self.__web_dir = web_dir
        
    def index(self):
        config, postinfo, bytespersec = build_header()
        
        config['download_dir'] = sabnzbd.CFG['misc']['download_dir']
        config['complete_dir'] = sabnzbd.CFG['misc']['complete_dir']
        config['cache_dir'] = sabnzbd.CFG['misc']['cache_dir']
        config['log_dir'] = sabnzbd.CFG['misc']['log_dir']
        config['nzb_backup_dir'] = sabnzbd.CFG['misc']['nzb_backup_dir']
        config['web_dir'] = sabnzbd.CFG['misc']['web_dir']
        config['dirscan_dir'] = sabnzbd.CFG['misc']['dirscan_dir']
            
        template = Template(file=os.path.join(self.__web_dir, 'config_directories.tmpl'),
                            searchList=[config],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True

    def saveDirectories(self, download_dir = None, complete_dir = None, log_dir = None,
                        cache_dir = None, web_dir = None, nzb_backup_dir = None,
                        dirscan_dir = None):
                        
        if download_dir and not os.access(os.path.abspath(download_dir), os.R_OK + os.W_OK):
            return "Error: can't access download directory."
        if not download_dir:
            return "Error: download directory not set."
            
        if cache_dir and not os.access(os.path.abspath(cache_dir), os.R_OK + os.W_OK):
            return "Error: can't access cache directory."
        if not cache_dir:
            return "Error: cache directory not set."
            
        if web_dir and not os.access(os.path.abspath(web_dir), os.R_OK):
            return "Error: can't access template directory."
        if not web_dir:
            return "Error: template directory not set."
            
        if log_dir and not os.access(os.path.abspath(log_dir), os.R_OK + os.W_OK):
            return "Error: can't access log directory."
            
        if dirscan_dir and not os.access(os.path.abspath(dirscan_dir), os.R_OK + os.W_OK):
            return "Error: can't access dirscan directory."
            
        if complete_dir and not os.access(os.path.abspath(complete_dir), os.R_OK + os.W_OK):
            return "Error: can't access complete directory."
            
        if nzb_backup_dir and not os.access(os.path.abspath(nzb_backup_dir), os.R_OK + os.W_OK):
            return "Error: can't access complete directory."
            
            
        sabnzbd.CFG['misc']['download_dir'] = download_dir        
        sabnzbd.CFG['misc']['cache_dir'] = cache_dir
        sabnzbd.CFG['misc']['web_dir'] = web_dir
        sabnzbd.CFG['misc']['log_dir'] = log_dir                
        sabnzbd.CFG['misc']['dirscan_dir'] = dirscan_dir
        sabnzbd.CFG['misc']['complete_dir'] = complete_dir
        sabnzbd.CFG['misc']['nzb_backup_dir'] = nzb_backup_dir
        return saveAndRestart(self.__root)
    saveDirectories.exposed = True
    
#------------------------------------------------------------------------------

class ConfigSwitches:
    def __init__(self, web_dir):
        self.__root = '/sabnzbd/config/switches'
        
        self.__web_dir = web_dir
        
    def index(self):
        config, postinfo, bytespersec = build_header()
        
        config['enable_unrar'] = int(sabnzbd.CFG['misc']['enable_unrar'])
        config['enable_unzip'] = int(sabnzbd.CFG['misc']['enable_unzip'])
        config['enable_filejoin'] = int(sabnzbd.CFG['misc']['enable_filejoin'])
        config['enable_save'] = int(sabnzbd.CFG['misc']['enable_save'])
        config['enable_par_cleanup'] = int(sabnzbd.CFG['misc']['enable_par_cleanup'])
        config['ram_cache'] = int(sabnzbd.CFG['misc']['ram_cache'])
        config['fail_on_crc'] = int(sabnzbd.CFG['misc']['fail_on_crc'])
        config['create_group_folders'] = int(sabnzbd.CFG['misc']['create_group_folders'])
        config['conserve_memory'] = int(sabnzbd.CFG['misc']['conserve_memory'])
        config['dirscan_opts'] = int(sabnzbd.CFG['misc']['dirscan_opts'])
        config['f_mode'] = int(sabnzbd.CFG['misc']['f_mode'])
        config['auto_sort'] = int(sabnzbd.CFG['misc']['auto_sort'])
        
        template = Template(file=os.path.join(self.__web_dir, 'config_switches.tmpl'),
                            searchList=[config],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True
    
    def saveSwitches(self, enable_unrar = None, enable_unzip = None,
                     enable_filejoin = None, enable_save = None,
                     ram_cache = None, fail_on_crc = None, f_mode = None,
                     create_group_folders = None, dirscan_opts = None,
                     conserve_memory = None, enable_par_cleanup = None,
                     auto_sort = None):
                     
        sabnzbd.CFG['misc']['enable_unrar'] = int(enable_unrar)
        sabnzbd.CFG['misc']['enable_unzip'] = int(enable_unzip)
        sabnzbd.CFG['misc']['enable_filejoin'] = int(enable_filejoin)
        sabnzbd.CFG['misc']['enable_save'] = int(enable_save)
        sabnzbd.CFG['misc']['ram_cache'] = int(ram_cache)
        sabnzbd.CFG['misc']['fail_on_crc'] = int(fail_on_crc)
        sabnzbd.CFG['misc']['create_group_folders'] = int(create_group_folders)
        sabnzbd.CFG['misc']['conserve_memory'] = int(conserve_memory)
        sabnzbd.CFG['misc']['dirscan_opts'] = int(dirscan_opts)
        sabnzbd.CFG['misc']['enable_par_cleanup'] = int(enable_par_cleanup)
        sabnzbd.CFG['misc']['f_mode'] = int(f_mode)
        sabnzbd.CFG['misc']['auto_sort'] = int(auto_sort)
        
        return saveAndRestart(self.__root)
    saveSwitches.exposed = True
    
#------------------------------------------------------------------------------

class ConfigGeneral:
    def __init__(self, web_dir):
        self.__root = '/sabnzbd/config/general'
        
        self.__web_dir = web_dir
        
    def index(self):
        config, postinfo, bytespersec = build_header()
        
        config['host'] = sabnzbd.CFG['misc']['host']
        config['port'] = sabnzbd.CFG['misc']['port']
        config['username'] = sabnzbd.CFG['misc']['username']
        config['password'] = sabnzbd.CFG['misc']['password']
        config['bandwith_limit'] = sabnzbd.CFG['misc']['bandwith_limit']
        config['username_newzbin'] = sabnzbd.CFG['newzbin']['username']
        config['password_newzbin'] = sabnzbd.CFG['newzbin']['password']
        
        if not sabnzbd.CFG['misc']['cleanup_list']:
            config['cleanup_list'] = ','
        
        elif len(sabnzbd.CFG['misc']['cleanup_list']) == 1:
            config['cleanup_list'] = '%s,' % sabnzbd.CFG['misc']['cleanup_list'][0]
        
        else:
            config['cleanup_list'] = listquote.makelist(sabnzbd.CFG['misc']['cleanup_list'])
        
        template = Template(file=os.path.join(self.__web_dir, 'config_general.tmpl'),
                            searchList=[config],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True
    
    def saveGeneral(self, host = None, port = None, username = None, password = None,
                    cronlines = None, username_newzbin = None, password_newzbin = None, 
                    bandwith_limit = None, cleanup_list = None):
        sabnzbd.CFG['misc']['host'] = host
        sabnzbd.CFG['misc']['port'] = port
        sabnzbd.CFG['misc']['username'] = username
        sabnzbd.CFG['misc']['password'] = password
        sabnzbd.CFG['misc']['bandwith_limit'] = bandwith_limit
        sabnzbd.CFG['newzbin']['username'] = username_newzbin
        sabnzbd.CFG['newzbin']['password'] = password_newzbin
        sabnzbd.CFG['misc']['cleanup_list'] = listquote.simplelist(cleanup_list)
        
        return saveAndRestart(self.__root)
    saveGeneral.exposed = True
    
#------------------------------------------------------------------------------

class ConfigServer:
    def __init__(self, web_dir):
        self.__root = '/sabnzbd/config/server'
        
        self.__web_dir = web_dir
        
    def index(self):
        config, postinfo, bytespersec = build_header()
        
        config['servers'] = sabnzbd.CFG['servers']
        
        template = Template(file=os.path.join(self.__web_dir, 'config_server.tmpl'),
                            searchList=[config],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True
    
    def saveServer(self, server = None, host = None, port = None, username = None,
                         password = None, connections = None, fillserver = None):
        
        if host and port and port.isdigit() and connections \
        and connections.isdigit() and fillserver and fillserver.isdigit():
            if server not in sabnzbd.CFG['servers']:
                server = host + ":" + port
                sabnzbd.CFG['servers'][server] = {}
            sabnzbd.CFG['servers'][server]['host'] = host
            sabnzbd.CFG['servers'][server]['port'] = port
            sabnzbd.CFG['servers'][server]['username'] = username
            sabnzbd.CFG['servers'][server]['password'] = password
            sabnzbd.CFG['servers'][server]['connections'] = connections
            sabnzbd.CFG['servers'][server]['fillserver'] = fillserver
            return saveAndRestart(self.__root)
        else:
            raise cherrypy.HTTPRedirect(self.__root)
    saveServer.exposed = True
    
    def delServer(self, server = None):
        del sabnzbd.CFG['servers'][server]
        return saveAndRestart(self.__root)
    delServer.exposed = True
    
class ConfigScheduling:
    def __init__(self, web_dir):
        self.__root = '/sabnzbd/config/scheduling'
        
        self.__web_dir = web_dir
        
    def index(self):
        config, postinfo, bytespersec = build_header()
        
        config['cronlines'] = sabnzbd.CFG['misc']['cronlines']
        
        template = Template(file=os.path.join(self.__web_dir, 'config_scheduling.tmpl'),
                            searchList=[config],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True
    
    def addSchedule(self, minute = None, hour = None, dayofmonth = None, month = None,
                    dayofweek = None, action = None):
                    
        if minute and hour and dayofmonth and month and dayofweek and action:
            sabnzbd.CFG['misc']['cronlines'].append('%s %s %s %s %s %s' %
                           (minute, hour, dayofmonth, month, dayofweek, action))
        return saveAndRestart(self.__root)
    addSchedule.exposed = True
    
    def delSchedule(self, line = None):
        print line
        if line and line in sabnzbd.CFG['misc']['cronlines']:
            sabnzbd.CFG['misc']['cronlines'].remove(line)
        return saveAndRestart(self.__root)
    delSchedule.exposed = True
    
class ConnectionInfo:
    def __init__(self, web_dir):
        self.__root = '/sabnzbd/connections'
        
        self.__web_dir = web_dir
        
    def index(self):
        header, postinfo, bytespersec = build_header()
        
        header['servers'] = []
        
        for server in sabnzbd.DOWNLOADER.servers[:]:
            connected = 0
            
            for nw in server.idle_threads[:]:
                if nw.connected:
                    connected += 1
                    
            for nw in server.busy_threads[:]:
                if nw.connected:
                    connected += 1
            
            header['servers'].append((server.host, server.port, connected))
            
        template = Template(file=os.path.join(self.__web_dir, 'connection_info.tmpl'),
                            searchList=[header],
                            compilerSettings={'directiveStartToken': '<!--#', 
                                              'directiveEndToken': '#-->'})
        return template.respond()
    index.exposed = True
    
    def disconnect(self):
        sabnzbd.disconnect()
        
        raise cherrypy.HTTPRedirect(self.__root)
    disconnect.exposed = True
    
def saveAndRestart(redirect_root):
    if type(sabnzbd.CFG.filename) is type(''):
        fn = os.path.abspath(sabnzbd.CFG.filename)
    else:
        fn = os.path.abspath(sabnzbd.CFG.filename.name)
    sabnzbd.CFG.filename = open(fn, 'w')
    sabnzbd.CFG.write()
    sabnzbd.CFG.filename.flush()
    sabnzbd.CFG = ConfigObj(fn)
    sabnzbd.halt()
    init_ok = sabnzbd.initialize()
    if init_ok:
        sabnzbd.start()
        raise cherrypy.HTTPRedirect(redirect_root)
    else:
        return "SABnzbd restart failed! See logfile(s)."
    
def build_header():
    try:
        uptime = calc_age(sabnzbd.START)
    except:
        uptime = "Error"
    
    header = { 'version':sabnzbd.__version__, 'paused':sabnzbd.paused(),
               'uptime':uptime }
    header['diskspace1'] = "%.2f" % diskfree(sabnzbd.DOWNLOAD_DIR)
    header['diskspace2'] = "%.2f" % diskfree(sabnzbd.COMPLETE_DIR)
    
    bytespersec = sabnzbd.bps()
    bytes, bytes_left, postinfo = sabnzbd.queue_info()
        
    header['kbpersec'] = "%.2f" % (bytespersec/1024.0)
    header['mbleft']   = "%.2f" % (bytes_left/1048576.0)
    header['mb']       = "%.2f" % (bytes/1048576.0)
    
    return (header, postinfo, bytespersec)
    
def calc_age(date):
    try:
        now = datetime.datetime.now()
        
        age = str(now - date).split(".")[0]
    except:
        age = "Error"
        
    return age
