# Copyright 2005 Gregor Kaufmann <tdn@gmx.at>
# Licensed under the Academic Free License 2.1
"""
sabnzbd.nzbqueue - nzb queue
"""

__NAME__ = "sabnzbd.nzbqueue"

import cPickle
import os
import re
import binascii
import logging
import tempfile
import sabnzbd
import time
import datetime

try:
    import _yenc
    HAVE_YENC = True

except ImportError:
    HAVE_YENC = False

#------------------------------------------------------------------------------

class FutureNzb:
    def __init__(self, postid, msg, repair, unpack, delete):
        self.postid = postid
        self.__filename = msg
        self.__repair = repair
        self.__unpack = unpack
        self.__delete = delete
        self.nzo_id = None
        
        self.futuretype = True
        
        self.deleted = False
        
        self.ram_cached = True
        
        self.parsed = True
        
        self.try_list = []
        
    def get_article(self, host, fillserver):
        return None
        
    def get_repair_opts(self):
        return (self.__repair, self.__unpack, self.__delete)
        
    def get_filename(self):
        return self.__filename
    
    def get_unpackstrht(self):
        return None
        
    def get_avg_date(self):
        return datetime.datetime.now()
        
    def gather_info(self):
        return (self.__repair, self.__unpack, self.__delete, self.__filename,
                (), 0, 0, None, (), (), ())
                
    def cleanup(self):
        pass
        
    def add_to_try_list(self, host):
        if host not in self.try_list:
            logging.debug("[%s] Appending %s to nzo.try_list (%s)",
                          __NAME__, host, self.__filename)
            self.try_list.append(host)
            
    def remove_from_try_list(self, host):
        if host in self.try_list:
            logging.debug("[%s] Removing %s from nzo.try_list (%s)", 
                          __NAME__, host, self.__filename)
                          
    def reset_try_list(self):
        if self.try_list:
            logging.info("[%s] Reseting nzo.try_list (%s)", __NAME__, 
                         self.__filename)
            self.try_list = []
            
    def get_all_files(self):
        return []
            
class HistoryItem:
    def __init__(self, nzo):
        self.nzo = nzo
        self.filename = nzo.get_filename()
        self.bytes_downloaded = nzo.bytes_downloaded
        self.completed = time.time()
        self.unpackstrht = None
        
    def __setstate__(self, _dict):
        self.__dict__.update(_dict)
        self.cleanup()
        
    def cleanup(self):
        if self.nzo:
            self.bytes_downloaded = self.nzo.bytes_downloaded
            self.unpackstrht = self.nzo.get_unpackstrht()
            self.completed = time.time()
            self.nzo = None
            
class CrcError(Exception):
    def __init__(self, needcrc, gotcrc):
        Exception.__init__(self)
        self.needcrc = needcrc
        self.gotcrc = gotcrc
        
YDEC_TRANS = ''.join([chr((i + 256 - 42) % 256) for i in range(256)])
class NzbQueue:
    def __init__(self, cache_file, cache_dir_items, cache_dir_articles, f_mode,
                 auto_sort):
        self.cache_file = cache_file
        self.cache_dir_items = cache_dir_items
        self.cache_dir_articles = cache_dir_articles
        
        self.downloaded_items = []
        
        self.items = []
        self.items_table = {}
        
        self.try_list = []
        
        self.bytes_downloaded = 0
        
        self.nzf_table = {}
        
        self.f_mode = f_mode
        
        self.auto_sort = auto_sort
        
        nzo_ids = []
        
        ## Readd old items
        if os.path.exists(self.cache_file):
            try:
                _f = open(self.cache_file, 'rb')
                (queue_vers, nzo_ids, self.downloaded_items) = cPickle.load(_f)
                _f.close()
                
                if not queue_vers == sabnzbd.__queueversion__:
                    logging.info("[%s] Outdated savefile found, discarding", 
                                 __NAME__)
                    self.downloaded_items = []
                    nzo_ids = []
            except:
                try:
                    _f.close()
                except:
                    pass
                logging.exception("[%s] Error loading %s, corrupt file " + \
                                       "detected", __NAME__, self.cache_file)
            for nzo_id in nzo_ids:
                path = os.path.join(self.cache_dir_items, nzo_id)
                if os.path.exists(path):
                    try:
                        _f = open(path, 'rb')
                        item = cPickle.load(_f)
                        _f.close()
                        self.add(item)
                    except EOFError:
                        try:
                            _f.close()
                        except:
                            pass
                        try:
                            os.remove(_f)
                        except:
                            pass
                        logging.exception("[%s] Error loading %s, corrupt " + \
                                          "file detected", __NAME__, path)
        # Remove unused articles
        if not self.items:
            try:
                dircontent = os.listdir(self.cache_dir_articles)
            except OSError:
                pass
            else:
                for article in dircontent:
                    path = os.path.join(self.cache_dir_articles, article)
                    if os.path.exists(path):
                        logging.info("[%s] Removing unused article: %s", 
                                     __NAME__, path)
                        try:
                            os.remove(path)
                        except OSError:
                            logging.exception("[%s] Error deleting: %s", 
                                              __NAME__, path)
    def save(self):
        """ Save queue """
        nzo_ids = []
        # Save all nzo_ids
        for item in self.items:
            nzo_ids.append(item.nzo_id)
            path = os.path.join(self.cache_dir_items, item.nzo_id)
            try:
                _f = open(path, 'wb')
                cPickle.dump(item, _f, 2)
                _f.flush()
                _f.close()
            except IOError:
                logging.exception("[%s] ERROR writing %s", __NAME__, path)
        try:
            _f = open(self.cache_file, 'wb')
            cPickle.dump((sabnzbd.__queueversion__, 
                          nzo_ids, self.downloaded_items), _f, 2)
            _f.flush()
            _f.close()
        except IOError:
            logging.exception("[%s] ERROR writing %s", __NAME__, 
                              self.cache_file)
                              
    def generate_future(self, article, msg, repair, unpack, delete):
        """ Create and return a FutureNzb Object """
        future = FutureNzb(article, msg, repair, unpack, delete)
        self.add(future)
        return future
        
    def insert_future(self, future, nzo):
        """ Replace a FutureNzb Object with a NzbObject Object """
        for i in xrange(len(self.items)):
            if future.nzo_id == self.items[i].nzo_id:
                # Get nzo_id and position from future
                nzo.nzo_id = future.nzo_id
                
                # Delete future
                logging.debug("[%s] Deleting future: %s", __NAME__, 
                              future.nzo_id)
                self.remove(future.nzo_id, False)
                
                # Insert NzbObject in place of future
                logging.debug("[%s] Inserting item: %s", __NAME__, nzo.nzo_id)
                self.add(nzo, i)
                break
                
    def add(self, nzo, pos = -1):
        # Reset try_lists
        self.reset_try_list()
        nzo.reset_try_list()
        
        # Need to generate a random id and filename => mkstemp
        if not nzo.nzo_id:
            try:
                fd, l = tempfile.mkstemp('', 'SABnzbd_', self.cache_dir_items)
                os.close(fd)
                head, tail = os.path.split(l)
                nzo.nzo_id = tail
            except IOError, (errno, strerror):
                logging.exception("[%s] ???", __NAME__)
                
        if nzo.nzo_id:
            nzo.deleted = False
            self.items_table[nzo.nzo_id] = nzo
            if pos > -1:
                self.items.insert(pos, nzo)
            else:
                self.items.append(nzo)
            self.save()
            
            ## file_mode stuff ##
            if self.f_mode:
                ## File mode won't work with conserve_memory
                if not nzo.parsed:
                    self.force_parse(nzo)
                        
                ## nzo possibly removed in force_parse
                if not nzo.deleted:
                    for nzf in nzo.get_all_files():
                        date = nzf.get_date()
                        if date not in self.nzf_table:
                            self.nzf_table[date] = []
                        self.nzf_table[date].append(nzf)
            ## file_mode stuff ##
        
        if self.auto_sort:
            self.sort_by_avg_age()
            
    def remove(self, nzo_id, add_to_history=True):
        if nzo_id in self.items_table:
            nzo = self.items_table[nzo_id]
            nzo.deleted = True
            self.items.remove(nzo)
            self.items_table.pop(nzo_id)
            
            for nzf in nzo.get_all_files():
                self.remove_from_nzf_table(nzf)
                
            if add_to_history:
                # Make sure item is only represented once in history
                should_add = True
                for hist_item in self.downloaded_items:
                    if hist_item.nzo and hist_item.nzo.nzo_id == nzo.nzo_id:
                        should_add = False
                        break
                if should_add:
                    self.downloaded_items.append(HistoryItem(nzo))
            else:
                self.cleanup_articles(nzo)
                
            path = os.path.join(self.cache_dir_items, nzo_id)
            if os.path.exists(path):
                logging.info("[%s] Removing %s", __NAME__, path)
                try:
                    os.remove(path)
                except OSError:
                     pass
                
    def remove_nzf(self, nzo_id, nzf_id):
        if nzo_id in self.items_table:    
            nzo = self.items_table[nzo_id]
            nzf = nzo.get_nzf_by_id(nzf_id)
            
            if nzf:
                self.remove_from_nzf_table(nzf)
                
                post_done = nzo.remove_nzf(nzf)
                if post_done:
                    self.remove(nzo_id, add_to_history = False)
                    
    def remove_from_nzf_table(self, nzf):
        date = nzf.get_date()
        if date in self.nzf_table:
            nzf_list = self.nzf_table[date]
            if nzf in nzf_list:
                logging.debug("[%s] Removing %s from nzf_table[%s]", 
                             __NAME__, nzf, date)
                nzf_list.remove(nzf)
            if not nzf_list:
                logging.debug("[%s] Removing %s from nzf_table", 
                             __NAME__, date)
                self.nzf_table.pop(date)
                
    def switch(self, item_id_1, item_id_2):
        item_id_pos1 = -1
        item_id_pos2 = -1
        for i in xrange(len(self.items)):
            if item_id_1 == self.items[i].nzo_id:
                item_id_pos1 = i
            elif item_id_2 == self.items[i].nzo_id:
                item_id_pos2 = i
            if (item_id_pos1 > -1) and (item_id_pos2 > -1):
                item = self.items[item_id_pos1]
                del self.items[item_id_pos1]
                self.items.insert(item_id_pos2, item)
                break
                
    def sort_by_avg_age(self):
        logging.info("[%s] Sorting by average date...", __NAME__)
        self.items.sort(cmp=nzo_date_cmp)
        
    def add_to_try_list(self, host):
        if host not in self.try_list:
            logging.debug("[%s] Appending %s to NZBQ.try_list",
                          __NAME__, host)
            self.try_list.append(host)
                
    def remove_from_try_list(self, host):
        if host in self.try_list:
            logging.debug("[%s] Removing %s from NZBQ.try_list", 
                          __NAME__, host)
            self.try_list.remove(host)
            
    def reset_try_list(self):
        if self.try_list:
            logging.debug("[%s] Reseting NZBQ.try_list", __NAME__)
            self.try_list = []
            
    def get_article(self, host, fillserver):
        if self.f_mode:
            keys = self.nzf_table.keys()
            keys.sort()
            
            for key in keys:
                for nzf in self.nzf_table[key]:
                    if (host not in nzf.try_list) and (nzf.is_active()):
                        article = nzf.get_article(host, fillserver)
                        if article:
                            return article
        else:
            for nzo in self.items:
                if not nzo.parsed:
                    self.force_parse(nzo)
                    
                    ## nzo possibly removed in force_parse
                    if nzo.deleted:
                        continue                        
                        
                # Don't try to get an article if host is in nzo.try_list
                if host not in nzo.try_list:
                    article = nzo.get_article(host, fillserver)
                    if article:
                        return article
                        
        # No articles for this host, block host (until reset issued)
        self.add_to_try_list(host)
        
    def register_article(self, article, lines = None):
        nzf = article.nzf
        nzo = nzf.nzo
        
        if nzo.deleted:
            logging.debug("[%s] Discarding article %s, no longer in queue", 
                          __NAME__, article.article)
            return
        
        if lines:
            try:
                self.decode(article, lines)
            except IOError, e:
                raise e
            except CrcError, e:
                raise e
            except:
                logging.exception("[%s] Unknown exception in decode()",
                                  __NAME__)
        
        file_done, post_done, reset = nzo.remove_article(article)
        
        if reset:
            self.reset_try_list()
            
        if file_done:
            self.remove_from_nzf_table(nzf)
                
            if sabnzbd.DO_SAVE:
                logging.info("[%s] Saving state [%s]", __NAME__, 
                                                            nzo.nzo_id)
                _f = open(os.path.join(self.cache_dir_items, nzo.nzo_id), 
                          'wb')
                cPickle.dump(nzo, _f, 2)
                _f.flush()
                _f.close()
                
            filename = nzf.get_filename()
            _type = nzf.get_type()
            
            # Only start decoding if we have a filename and type
            if filename and _type:
                sabnzbd.assemble_nzf((nzo, nzf))
                
            else:
                logging.warning('[%s] %s -> Unknown encoding', __NAME__,
                                filename)
        
        if post_done:
            self.remove(nzo.nzo_id, True)
            
            # Notify assembler to call postprocessor
            sabnzbd.assemble_nzf((nzo, None))
            
    def purge(self):
        self.downloaded_items = []
        
    def queue_info(self):
        bytes_left = 0
        bytes = 0
        postinfo = []
        for nzo in self.items:
            nfo = nzo.gather_info()
            
            bytes += nfo[6] 
            bytes_left += nfo[5]
            postinfo.append((nzo.nzo_id, nfo))
                                     
        return (bytes, bytes_left, postinfo)
        
    def history_info(self):
        history_info = {}
        bytes_downloaded = 0
        for hist_item in self.downloaded_items:
            completed = hist_item.completed
            filename = hist_item.filename
            bytes_downloaded += hist_item.bytes_downloaded
            
            if completed not in history_info:
                history_info[completed] = []
                
            if hist_item.nzo:
                unpackstrht = hist_item.nzo.get_unpackstrht()
                loaded = True
            else:
                unpackstrht = hist_item.unpackstrht
                loaded = False
                
            history_info[completed].append((filename, unpackstrht, loaded))
        return (history_info, bytes_downloaded)
        
    def is_empty(self):
        empty = True
        for nzo in self.items:
            if not nzo.futuretype:
                empty = False
                break
        return empty
        
    def cleanup_nzo(self, nzo):
        
        self.cleanup_articles(nzo)
        
        for hist_item in self.downloaded_items:
            # refresh fields & delete nzo reference
            if hist_item.nzo and hist_item.nzo == nzo:
                hist_item.cleanup()
                logging.debug('[%s] %s cleaned up', __NAME__, 
                nzo.get_filename())
                
    def cleanup_articles(self, nzo):
        if not nzo.ram_cached:
            for article in nzo.saved_articles:
                path = os.path.join(self.cache_dir_articles, article) 
                if os.path.exists(path):
                    try:
                        logging.info("[%s] Removing unused article: %s", 
                                     __NAME__, path)
                        os.remove(path)
                    except:
                        logging.exception("[%s] Error deleting: %s", 
                                          __NAME__, path)
                                          
    def decode(self, article, data):
        data = strip(data)
        ## No point in continuing if we don't have any data left
        if data:
            nzf = article.nzf
            nzo = nzf.nzo
            yenc, data = yCheck(data)
            ybegin, ypart, yend = yenc
            
            #Deal with non-yencoded posts
            if not ybegin:
                if data[0].startswith('begin '):
                    nzf.set_filename(data[0].split(None, 2)[2])
                    nzf.set_type('uu')
                    data.pop(0)
                if data[-1] == 'end':
                    data.pop()
                    if data[-1] == '`':
                        data.pop()
                        
                if nzo.ram_cached:
                    nzf.add_data(article, data)
                else:
                    path = os.path.join(self.cache_dir_articles, article.article)
                    try:
                        f_tmp = open(path, 'wb')
                        cPickle.dump(data, f_tmp, 2)
                        f_tmp.close()
                        nzo.saved_articles.append(article.article)
                    except IOError, (errno, strerror):
                        try:
                            f_tmp.close()
                        except:
                            pass
                        raise
                    except:
                        try:
                            f_tmp.close()
                        except:
                            pass
                        logging.exception('[%s] Unknown exception', __NAME__)
                        
            #Deal with yenc encoded posts
            elif (ybegin and yend):
                if not nzf.get_filename():
                    if 'name' in ybegin:
                        nzf.set_filename(ybegin['name'])
                    else:
                        logging.debug("[%s] Possible corrupt header detected " + \
                                      "=> ybegin: %s", __NAME__, ybegin)
                if not nzf.get_type():
                    nzf.set_type('yenc')
                # Decode data
                if HAVE_YENC:
                    decoded_data, crc = _yenc.decode_string(''.join(data))[:2]
                    partcrc = '%08X' % ((crc ^ -1) & 2**32L - 1)
                else:
                    data = ''.join(data)
                    for i in (0, 9, 10, 13, 27, 32, 46, 61):
                        j = '=%c' % (i + 64)
                        data = data.replace(j, chr(i))
                        decoded_data = data.translate(YDEC_TRANS)
                        crc = binascii.crc32(decoded_data)
                        partcrc = '%08X' % (crc & 2**32L - 1)
                        
                if ypart:
                    crcname = 'pcrc32'
                else:
                    crcname = 'crc32'
                    
                if crcname in yend:
                    _partcrc = '0' * (8 - len(yend[crcname])) + yend[crcname].upper()
                else:
                    _partcrc = None
                    logging.debug("[%s] Possible corrupt header detected " + \
                                  "=> yend: %s", __NAME__, yend)
                                  
                # Save data
                if nzo.ram_cached:
                    nzf.add_data(article, decoded_data)
                else:
                    path = os.path.join(self.cache_dir_articles, article.article)
                    try:
                        f_tmp = open(path, 'wb')
                        f_tmp.write(decoded_data)
                        f_tmp.close()
                        nzo.saved_articles.append(article.article)
                    except IOError:
                        try:
                            f_tmp.close()
                        except:
                            pass
                        raise
                    except:
                        try:
                            f_tmp.close()
                        except:
                            pass
                        logging.exception('[%s] Unknown exception', __NAME__)
                        
                if not (_partcrc == partcrc):
                    raise CrcError(_partcrc, partcrc)
                    
    def force_parse(self, nzo):
        filename = nzo.get_filename()
        
        logging.debug("[%s] Forcing parse of %s", __NAME__, filename)
        try:
            nzo.parse()
        except:
            logging.exception("[%s] Parsing %s failed", __NAME__, filename)
            self.remove(nzo.nzo_id, add_to_history = False)
            
#------------------------------------------------------------------------------
def yCheck(data):
    ybegin = None
    ypart = None
    yend = None
    
    ## Check head
    for i in xrange(10):
        try:
            if data[i].startswith('=ybegin '):
                ybegin = ySplit(data[i])
                if data[i+1].startswith('=ypart '):
                    ypart = ySplit(data[i+1])
                    data = data[i+2:]
                    break
                else:
                    data = data[i+1:]
                    break
        except IndexError:
            break
    
    ## Check tail
    for i in xrange(-1, -11, -1):
        try:
            if data[i].startswith('=yend '):
                yend = ySplit(data[i])
                data = data[:i]
                break
        except IndexError:
            break
            
    return ((ybegin, ypart, yend), data)
    
def ySplit(line):
    YSPLIT_RE = re.compile(r'(\S+)=')
    fields = {}
    
    parts = YSPLIT_RE.split(line)[1:]
    if len(parts) % 2:
        return fields
        
    for i in range(0, len(parts), 2):
        key, value = parts[i], parts[i+1]
        fields[key] = value.strip()
        
    return fields
    
def strip(data):
    tmp = None
    while data and not tmp:
        tmp = data.pop()
        if tmp:
            data.append(tmp)
            break
            
    tmp = None
    while not tmp and data:
        tmp = data.pop(0)
        if tmp:
            data.insert(0, tmp)
            break
            
    for i in xrange(len(data)):
        if data[i][:2] == '..':
            data[i] = data[i][1:]
    return data
    
def nzo_date_cmp(nzo1, nzo2):
    return cmp(nzo1.get_avg_date(), nzo2.get_avg_date())
