# Copyright 2005 Gregor Kaufmann <tdn@gmx.at>
# Licensed under the Academic Free License 2.1

"""
sabnzbd.nzbstuff - misc
"""
import re
import os
import logging
import sabnzbd
import tempfile
import cPickle
import datetime
import bz2

try:
    from cElementTree import XML
    HAVE_CELEMENTTREE = True
except ImportError:
    from elementtree.ElementTree import XML
    HAVE_CELEMENTTREE = False
        
__NAME__ = "sabnzbd.nzbstuff"

#------------------------------------------------------------------------------
class Article:
    def __init__ (self, article, bytes, partnum, nzf):
        self.fetched = False
        self.allow_fill_server = False
        
        self.article = article
        self.bytes = bytes
        self.partnum = partnum
        self.nzf = nzf
        
        self.try_list = []
        
    def get_article(self, host, fillserver):
        if fillserver and not self.allow_fill_server:
            return None
            
        if not self.fetched and host not in self.try_list:
            self.fetched = True
            self.try_list.append(host)
            return self
            
    def __setstate__(self, _dict):
        self.__dict__.update(_dict)
        self.fetched = False
        self.try_list = []
        
class NzbFile:
    def __init__(self, date, subject, segments, nzo):
        # Private
        self.__date = date
        self.__subject = subject
        self.__filename = None
        self.__type = None
        
        self.__ispar2file = False
        self.__vol = None
        self.__blocks = None
        self.__setname = None
        self.__extrapars = None
        
        self.__initial_article = None
        
        self.__articles = []
        self.__decodetable = {}
        self.__datatable = {}
        
        self.__bytes = 0
        self.__bytes_left = 0
        
        # Public
        self.nzo = nzo
        self.nzf_id = str(id(self))
        
        self.try_list = []
        
        for segment in segments:
            bytes = int(segment.get('bytes'))
            partnum = int(segment.get('number'))
            
            if partnum in self.__decodetable:
                logging.info("[%s] Skipping duplicate article (%s)", __NAME__,
                             segment.text)
                continue
                
            article = Article(segment.text, bytes, partnum, self)
            
            if partnum == 1:
                self.__initial_article = article
                
            self.__bytes += bytes
            self.__bytes_left += bytes
            self.__articles.append(article)
            self.__decodetable[partnum] = article
            self.__datatable[article] = None
            
    def bytes(self):
        return self.__bytes
        
    def bytes_left(self):
        return self.__bytes_left
        
    def remove_article(self, article):
        self.__articles.remove(article)
        self.__bytes_left -= article.bytes
        
        reset = False
        if article.partnum == 1 and self.__articles:
            # Issue reset
            self.__initial_article = None
            self.reset_try_list()
            reset = True
        
        done = True
        if self.__articles:
            done = False
            
        return (done, reset)
        
    def get_article(self, host, fillserver):
        if self.__initial_article:
            article = self.__initial_article.get_article(host, fillserver)
            if article:
                return article
                
        else:
            for article in self.__articles:
                article = article.get_article(host, fillserver)
                if article:
                    return article
                    
        # No articles for this host, block host (until reset issued)
        self.add_to_try_list(host)
    
    def is_active(self):
        return self.nzo.is_active_nzf(self)
        
    def get_subject(self):
        return self.__subject
        
    def get_date(self):
        return self.__date
        
    def get_filename(self):
        return self.__filename
        
    def get_decodetable(self):
        return self.__decodetable
        
    def get_type(self):
        return self.__type
        
    def get_extrapars(self):
        return self.__extrapars
        
    def get_blocks(self):
        return self.__blocks
        
    def get_setname(self):
        return self.__setname
        
    def set_type(self, _type):
        self.__type = _type
        
    def set_filename(self, filename):
        self.__filename = filename
        
    def set_par2(self, setname, vol, blocks):
        self.__ispar2file = True
        self.__setname = setname
        self.__vol = vol
        self.__blocks = int(blocks)
            
    def set_extrapars(self, extrapars):
        self.__extrapars = extrapars
            
    def remove_extrapar(self, extrapar):
        self.__extrapars.remove(extrapar)
        
    def add_data(self, article, data):
        self.__datatable[article] = data
        
    def get_datatable(self):
        return self.__datatable
        
    def add_to_try_list(self, host):
        if host not in self.try_list:
            logging.debug("[%s] Appending %s to nzf.try_list (%s)",
                          __NAME__, host, self.__subject)
            self.try_list.append(host)
            
    def remove_from_try_list(self, host):
        if host in self.try_list:
            logging.debug("[%s] Removing %s from nzf.try_list (%s)", 
                          __NAME__, host, self.__subject)
            self.try_list.remove(host)
            
    def reset_try_list(self):
        if self.try_list:
            logging.info("[%s] Reseting nzf.try_list (%s)", __NAME__, 
                         self.__filename)
            self.try_list = []
            
    def completed(self):
        return not bool(self.__articles)
            
    def __setstate__(self, _dict):
        self.__dict__.update(_dict)
        self.try_list = []
        
    def __repr__(self):
        return "<NzbFile: filename=%s, type=%s>" % \
               (self.__filename, self.__type)
               
NZBFN_MATCHER = re.compile(r"msgid_\d*_(.*).nzb", re.I)
PROBABLY_PAR2_RE = re.compile(r'(.*)\.vol(\d*)\+(\d*)\.par2', re.I)
class NzbObject:
    def __init__(self, filename, nzb, repair, unpack, delete):
        self.__filename = filename    # Original filename 
        self.__dirname = filename
        self.__bytes = 0              # Original bytesize
        self.__repair = repair        # True if we want to repair this set
        self.__unpack = unpack        # True if we want to unpack this set
        self.__delete = delete        # True if we want to delete this set
        self.__group = None
        
        self.__partable = {}
        self.__extrapars = {}
        
        self.__files = []
        self.__files_table = {}
        
        self.__finished_files = []
        
        self.__unpackstrht = {}
        
        self.__avg_date = None
        
        self.bytes_downloaded = 0
        
        self.saved_articles = []
        
        self.try_list = []
        
        self.nzo_id = None
        
        self.futuretype = False
        
        self.deleted = False
        
        self.parsed = False
        
        self.ram_cached = sabnzbd.RAM_CACHE
        self.create_group_folder = sabnzbd.CREATE_GROUP_FOLDERS
        
        self.nzb = None
        if  sabnzbd.CONSERVE_MEMORY:
            logging.debug("[%s] NzbObject: (bz2) Compressing nzb", __NAME__)
            self.nzb = bz2.compress(nzb)
        
        match_result = re.search(NZBFN_MATCHER, filename)
        if match_result:
            self.__dirname = match_result.group(1)
        
        self.__dirname = self.__dirname.replace('.', '_')
        
        if self.nzb:
            # Conserve memory -> just try to read in bytes
            try:
                for token in nzb.split():
                    if "bytes=" in token:
                        self.__bytes += int(token[7:-1])
            except:
                pass
        else:
            # Don't conserve memory -> parse now
            self.parse(nzb)
            
    def parse(self, nzb = None):
        self.parsed = True
        self.__bytes = 0
        
        blah = '{http://www.newzbin.com/DTD/2003/nzb}'
        if nzb:
            root = XML(nzb)
        else:
            logging.debug("[%s] NzbObject.parse(): decompressing nzb", __NAME__)
            root = XML(bz2.decompress(self.nzb))
            self.nzb = None
           
        avg_age = 0
        for _file in root:
            if not self.__group:
                groups = _file.find('%sgroups' % blah)
                if not groups:
                    groups = _file.find('groups')
                self.__group = groups[0].text
            subject = _file.get('subject')
            if isinstance(subject, unicode):
                subject = subject.encode('utf-8')
                
            t = int(_file.get('date'))
            date = datetime.datetime.fromtimestamp(t)
            
            avg_age += t
            
            segments = _file.find('%ssegments' % blah)
            if not segments:
                segments = _file.find('segments')
            nzf = NzbFile(date, subject, segments, self)
            
            self.__files_table[nzf.nzf_id] = nzf
            self.__bytes += nzf.bytes()
            valid = 1 in nzf.get_decodetable()
            
            if valid:
                logging.info('[%s] %s added to queue', __NAME__, subject)
                self.__files.append(nzf)
            else:
                logging.info('[%s] %s missing first segment, skipping file', 
                             __NAME__, subject)
                             
        self.__avg_date = datetime.datetime.fromtimestamp(avg_age / len(root))
        
        self.__files.sort(cmp=nzf_date_cmp)
    
    def remove_nzf(self, nzf):
        if nzf in self.__files:
            self.__files.remove(nzf)
            self.__finished_files.append(nzf)
        return not bool(self.__files)
        
    def get_nzf_by_id(self, nzf_id):
        if nzf_id in self.__files_table:
            nzf = self.__files_table[nzf_id]
            return nzf
        
    def is_active_nzf(self, nzf):
        return (nzf in self.__files) 
            
    def remove_article(self, article):
        nzf = article.nzf
        file_done, reset = nzf.remove_article(article)
        
        if file_done and (nzf in self.__files):
            self.__files.remove(nzf)
            self.__finished_files.append(nzf)
            
        if reset:
            self.reset_try_list()
            
        ## Special treatment for first part
        if article.partnum == 1:
            fn = nzf.get_filename()
            if fn:
                par2match = re.search(PROBABLY_PAR2_RE, fn)
                ## Is a par2file and repair mode activated
                if par2match and self.__repair:
                    head = par2match.group(1)
                    nzf.set_par2(par2match.group(1),
                                par2match.group(2),
                                par2match.group(3))
                    ## Already got a parfile for this set?
                    if head in self.__partable:
                        nzf.set_extrapars(self.__extrapars[head])
                        ## Set the smallest par2file as initialparfile
                        if nzf.get_blocks() < self.__partable[head].get_blocks() \
                        and self.__partable[head] in self.__files:
                            ## But only do this if our last initialparfile 
                            ## isn't already done (e.g two small parfiles)
                            self.__files.remove(self.__partable[head])
                            self.__extrapars[head].append(self.__partable[head])
                            self.__partable[head] = nzf
                            
                        ## This file either has more blocks,
                        ## or initialparfile is already decoded
                        else:
                            if not file_done:
                                self.__files.remove(nzf)
                                self.__extrapars[head].append(nzf)                                   
                    ## No par2file in this set yet, set this as 
                    ## initialparfile
                    else:
                        self.__partable[head] = nzf
                        self.__extrapars[head] = []
                        nzf.set_extrapars(self.__extrapars[head])
                ## Is not a par2file
                else:
                    pass
            ## No filename in seg 1? Probably not uu or yenc encoded
            ## Set subject as filename and skip
            else:
                nzf.set_filename(nzf.get_subject())
                if not file_done and (nzf in self.__files):
                    self.__files.remove(nzf)
                    self.__finished_files.append(nzf)
                    
        post_done = False
        if not self.__files:
            post_done = True
            
        return (file_done, post_done, reset)
        
    def set_unpackstr(self, msg, action, stage):
        if stage not in self.__unpackstrht:
            self.__unpackstrht[stage] = {}
        self.__unpackstrht[stage][action] = msg
        
    def add_parfile(self, parfile):
        self.__files.append(parfile)
        parfile.remove_extrapar(parfile)
        
    def remove_parset(self, setname):
        self.__partable.pop(setname)
        
    def get_unpackstrht(self):
        return self.__unpackstrht.copy()
        
    def get_repairsets(self):
        return self.__partable.keys()
        
    def get_repair_opts(self):
        return (self.__repair, self.__unpack, self.__delete)
            
    def get_partable_by_setname(self, set):
        if set in self.__partable:
            return self.__partable[set]
            
    def get_article(self, host, fillserver):
        for nzf in self.__files:
            # Don't try to get an article if host is in nzf.try_list
            if host not in nzf.try_list:
                article = nzf.get_article(host, fillserver)
                if article:
                    return article
        # No articles for this server, block for next time
        self.add_to_try_list(host)
        
    def get_dirname(self):
        return self.__dirname
        
    def get_filename(self):
        return self.__filename
        
    def get_group(self):
        return self.__group
        
    def get_all_files(self):
        files = []
        files.extend(self.__files)
        
        for _set in self.__extrapars:
            files.extend(self.__extrapars[_set])
            
        return files
        
    def get_avg_date(self):
        return self.__avg_date
        
    def bytes(self):
        return self.__bytes
        
    def bytes_left(self):
        bytes = 0
        for _file in self.__files:
            bytes += _file.bytes_left()
        return bytes
        
    def add_to_try_list(self, host):
        if host not in self.try_list:
            logging.debug("[%s] Appending %s to nzo.try_list (%s)",
                          __NAME__, host, self.__filename)
            self.try_list.append(host)
            
    def remove_from_try_list(self, host):
        if host in self.try_list:
            logging.debug("[%s] Removing %s from nzo.try_list (%s)", 
                          __NAME__, host, self.__filename)
            self.try_list.remove(host)
                          
    def reset_try_list(self):
        if self.try_list:
            logging.info("[%s] Reseting nzo.try_list (%s)", __NAME__, 
                         self.__filename)
            self.try_list = []
            
    def gather_info(self):
        if self.parsed:
            bytes_left_all = 0
        else:
            bytes_left_all = self.__bytes
            
        active_files = []
        queued_files = []
        finished_files = []
        
        for _file in self.__finished_files:
            bytes = _file.bytes()
            filename = _file.get_filename()
            date = _file.get_date()
            
            finished_files.append((0, bytes, filename, date))
            
        for _file in self.__files:
            bytes_left = _file.bytes_left()
            bytes = _file.bytes()
            filename = _file.get_filename()
            if not filename:
                filename = _file.get_subject()
            date = _file.get_date()
            
            bytes_left_all += bytes_left
            active_files.append((bytes_left, bytes, filename, date, 
                                 _file.nzf_id))
                                 
        for _set in self.__extrapars:
            for _file in self.__extrapars[_set]:
                bytes_left = _file.bytes_left()
                bytes = _file.bytes()
                filename = _file.get_filename()
                if not filename:
                    filename = _file.get_subject()
                date = _file.get_date()
                
                queued_files.append((_set, bytes_left, bytes, filename, date))
        
        return (self.__repair, self.__unpack, self.__delete, 
                self.__filename, self.__unpackstrht.copy(), 
                bytes_left_all, self.__bytes, self.__avg_date,
                finished_files, active_files, queued_files)
                                          
    def __setstate__(self, _dict):
        self.__dict__.update(_dict)
        self.try_list = []
        
def nzf_date_cmp(nzf1, nzf2):
    return cmp(nzf1.get_date(), nzf2.get_date())
